import unittest

import numpy as np
import pandas as pd
from dsbox.ml.explain.feature_contributions import FeatureContributions
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split


class FeatureContributionsTest(unittest.TestCase):
    def test_feature_contributions(self):
        # set random seed for test
        seed = 42
        np.random.seed(seed)

        X, y = make_classification(n_samples=30, n_features=3, n_redundant=0, n_informative=2, random_state=seed)

        X_train, X_test, y_train, y_test = \
            train_test_split(X, y, test_size=.2, random_state=seed)

        X_train = pd.DataFrame(X_train)
        X_test = pd.DataFrame(X_test)

        clf = RandomForestClassifier(n_estimators=10, random_state=seed)

        clf.fit(X_train, y_train)

        fc = FeatureContributions(clf)
        predictions, bias, contributions = fc.predict(X_test)

        expected_contributions = np.array([
            [[-0.00791667, 0.00791667],
             [-0.23905303, 0.23905303],
             [0.28863636, -0.28863636]],

            [[0.06117424, -0.06117424],
             [-0.06981061, 0.06981061],
             [-0.34969697, 0.34969697]],

            [[0.11117424, -0.11117424],
             [0.08018939, -0.08018939],
             [0.35030303, -0.35030303]],

            [[0.11117424, -0.11117424],
             [-0.01981061, 0.01981061],
             [-0.34969697, 0.34969697]],

            [[0.06117424, -0.06117424],
             [-0.06981061, 0.06981061],
             [-0.34969697, 0.34969697]],

            [[0.06117424, -0.06117424],
             [0.03018939, -0.03018939],
             [0.35030303, -0.35030303]]])

        contributions = np.round(contributions, decimals=7)
        expected_contributions = np.round(expected_contributions, decimals=7)

        self.assertEqual(np.array_equal(contributions, expected_contributions), True)



import collections

import numpy as np
from sklearn.metrics import mean_squared_error, precision_recall_curve, recall_score, precision_score

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def generalized_jaccard_similarity_score(y_true, y_pred):
    """ Generalized Jaccard similarity score

    The Generalized Jaccard similarity [1] is an extension of the Jaccard similarity coefficient.
    It handles multisets as input variables.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float

    References
    ----------
    . [1] `Wikipedia entry for the Jaccard index
           <https://en.wikipedia.org/wiki/Jaccard_index#Generalized_Jaccard_similarity_and_distance>`_

    """

    min_count = 0
    max_count = 0

    counter_y_true = collections.Counter(y_true)
    counter_y_pred = collections.Counter(y_pred)

    # union of all elements from y_true and y_pred
    elements = set(y_true).union(set(y_pred))

    for element in elements:
        max_count += max(counter_y_true[element], counter_y_pred[element])
        min_count += min(counter_y_true[element], counter_y_pred[element])

    # could happen only if both collections are empty
    if max_count == 0:
        return 0.0

    return min_count / float(max_count)


def root_mean_squared_error(y_true, y_pred):
    """ Root Mean Square Error

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float

    References
    ----------
    . [1] `Wikipedia entry for the Root-mean-square deviation
           <https://en.wikipedia.org/wiki/Root-mean-square_deviation>`

    """
    return np.sqrt(mean_squared_error(y_true, y_pred))


def mean_squared_percentage_error(y_true, y_pred):
    """ Mean Square Percentage Error

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float

    References
    ----------
    . [1] `Wikipedia entry for the Mean percentage error
           <https://en.wikipedia.org/wiki/Mean_percentage_error>`

    """
    return np.mean(np.ma.masked_invalid(np.square((y_true - y_pred) / y_true)))


def root_mean_squared_percentage_error(y_true, y_pred):
    """ Root Mean Square Percentage Error

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float

    References
    ----------
    . [1] `Wikipedia entry for the Mean percentage error
           <https://en.wikipedia.org/wiki/Mean_percentage_error>`

    """
    return np.sqrt(mean_squared_percentage_error(y_true, y_pred))


def mean_absolute_percentage_error(y_true, y_pred):
    """ Mean Absolute Percentage Error

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float

    References
    ----------
    . [1] `Wikipedia entry for the Mean absolute percentage error
           <https://en.wikipedia.org/wiki/Mean_absolute_percentage_error>`

    """
    return np.mean(np.ma.masked_invalid(np.abs((y_true - y_pred) / y_true)))


def symmetric_mean_absolute_percentage_error(y_true, y_pred):
    """ Symmetric Mean Absolute Percentage Error

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float
    """
    return np.mean(np.ma.masked_invalid(np.abs((y_true - y_pred) / ((y_true + y_pred) / 2))))


def symmetric_root_mean_squared_percentage_error(y_true, y_pred):
    """ Symetric Root Mean Square Percentage Error

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier.

    Returns
    -------
    score : float
    """
    return np.sqrt(np.mean(np.ma.masked_invalid(np.square((y_true - y_pred) / ((y_true + y_pred) / 2)))))


def precision_with_fixed_recall(y_true, y_pred_proba, fixed_recall):
    """ Compute precision with a fixed recall, for class 1. The chosen threshold for this couple precision is also returned.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred_proba : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier, should be probabilities

    fixed_recall : float
        Fixed recall, recall will be calculated with respect to this precision

    Returns
    -------
    precision_score : float
    threshold : float
    """
    if is_valid_y_true(y_true):
        _, recall, threshold = precision_recall_curve(y_true, y_pred_proba)
        threshold = max([threshold[i] for i in range(len(threshold)) if recall[i] >= fixed_recall])
        y_pred_binary = binarize(y_pred_proba, threshold)
        return precision_score(y_true, y_pred_binary), threshold
    else:
        raise ValueError('y_true should not contain only zeros')


def recall_with_fixed_precision(y_true, y_pred_proba, fixed_precision):
    """ Compute recall with a fixed precision, for class 1. The chosen threshold for this couple recall is also returned.

    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix
        Ground truth (correct) labels.

    y_pred_proba : 1d array-like, or label indicator array / sparse matrix
        Predicted labels, as returned by a classifier, should be probabilities

    fixed_precision : float
        Fixed precision, recall will be calculated with respect to this precision

    Returns
    -------
    recall_score : float
    threshold : float
    """
    if is_valid_y_true(y_true):
        precision, _, threshold = precision_recall_curve(y_true, y_pred_proba)
        threshold = min([threshold[i] for i in range(len(threshold)) if precision[i] >= fixed_precision])
        y_pred_binary = binarize(y_pred_proba, threshold)
        return recall_score(y_true, y_pred_binary), threshold
    else:
        raise ValueError('y_true should not contain only zeros')


def binarize(probas, threshold):
    """ Convert probabilities into 1s if superior or equal to the threshold, 0s otherwise.

    Parameters
    ----------
    probas : 1d array-like
        float probabilities to convert into binaries

    threshold : float

    Returns
    -------
    y_pred_binary : 1d array-like of 0s and 1s
    """
    return [1 if proba >= threshold else 0 for proba in probas]


def is_valid_y_true(y_true):
    """
    Return true if y_true contains at least one 1, false otherwise
    Parameters
    ----------
    y_true : 1d array-like, or label indicator array / sparse matrix

    Returns
    -------
    is_valid_y_true : bool
    """
    return np.sum(y_true) > 0



import warnings

import numpy as np
import pyarrow as pa
import pyarrow.plasma as plasma

__status__ = "Beta"


class PlasmaConnector:
    """
    Based on PyArrow documentation
    """

    def __init__(self, socket_name, manager_socket_name="", int_release_delay=0):
        warnings.warn("beta version", Warning)
        self.socket_name = socket_name
        self.manager_socket_name = manager_socket_name
        self.int_release_delay = int_release_delay

        self.client = plasma.connect(socket_name, manager_socket_name, int_release_delay)

    @staticmethod
    def generate_object_id():
        return plasma.ObjectID(np.random.bytes(20))

    def put_dataframe(self, dataframe, object_id=None, overwrite=True):
        if object_id is None:
            object_id = PlasmaConnector.generate_object_id()

        # Delete object if exists
        if overwrite:
            self.client.delete([object_id])
        else:
            if self.client.contains(object_id):
                raise ValueError("object id already exists.")

        # Convert the Pandas DataFrame into a PyArrow RecordBatch
        record_batch = pa.RecordBatch.from_pandas(dataframe)

        # Create the Plasma object from the PyArrow RecordBatch. Most of the work here
        # is done to determine the size of buffer to request from the object store.
        mock_sink = pa.MockOutputStream()
        stream_writer = pa.RecordBatchStreamWriter(mock_sink, record_batch.schema)
        stream_writer.write_batch(record_batch)
        stream_writer.close()

        data_size = mock_sink.size()
        buffer = self.client.create(object_id, data_size)

        # Write the PyArrow RecordBatch to Plasma
        stream = pa.FixedSizeBufferWriter(buffer)
        stream_writer = pa.RecordBatchStreamWriter(stream, record_batch.schema)
        stream_writer.write_batch(record_batch)
        stream_writer.close()

        # Seal the Plasma object (make it immutable and usable by others clients)
        self.client.seal(object_id)

        return object_id

    def get_dataframe(self, object_id):
        # Fetch the Plasma object
        [data] = self.client.get_buffers([object_id])  # Get PlasmaBuffer from ObjectID
        buffer = pa.BufferReader(data)

        # Convert object back into an Arrow RecordBatch
        reader = pa.RecordBatchStreamReader(buffer)
        record_batch = reader.read_next_batch()

        # Convert back into Pandas
        return record_batch.to_pandas()



from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults


class DataOperator(BaseOperator):
    """

    Data Operator is a custom operator which can handle input and/or output data.
    The parameters of the operation function has to be passed as a dict named 'params', based on BaseOperator behavior.

    Parameters
    ----------
    operation_function: function
        function taking the output of the input_unit and returning its result to the input of the output_unit
    input_unit: DataInputUnit, default=None
        define input data reading
    output_unit: DataOutputUnit, default=None
        define output data writing
    """

    ui_color = '#fff2e6'

    @apply_defaults
    def __init__(self, operation_function, input_unit=None, output_unit=None, *args, **kwargs):
        self.operation_function = operation_function
        self.input_unit = input_unit
        self.output_unit = output_unit

        super(DataOperator, self).__init__(*args, **kwargs)

    def execute(self, context):
        data = None
        if self.input_unit is not None:
            data = self.input_unit.read_data()
            data = self.operation_function(data, **self.params)
        else:
            data = self.operation_function(**self.params)
        if self.output_unit is not None:
            self.output_unit.write_data(data)



from tensorflow import keras
import numpy as np
import matplotlib.cm as cm


def superimposed_visualization(img_path, heatmap):
    """
    Used to understand the prediction of a model using a heatmap pixel contribution.

    Parameters
    ----------
    img_path: str
        path to image file
    heatmap: 2D numpy array
        heatmap pixel contribution

    Returns
    -------
    original image with superimposed heatmap

    """
    img = keras.preprocessing.image.load_img(img_path)
    img = keras.preprocessing.image.img_to_array(img)

    # We rescale heatmap to a range 0-255
    heatmap = np.uint8(255 * heatmap)

    # We use jet colormap to colorize heatmap
    jet = cm.get_cmap("jet")

    # We use RGB values of the colormap
    jet_colors = jet(np.arange(256))[:, :3]
    jet_heatmap = jet_colors[heatmap]

    # We create an image with RGB colorized heatmap
    jet_heatmap = keras.preprocessing.image.array_to_img(jet_heatmap)
    jet_heatmap = jet_heatmap.resize((img.shape[1], img.shape[0]))
    jet_heatmap = keras.preprocessing.image.img_to_array(jet_heatmap)

    # Superimpose the heatmap on original image
    superimposed_img = jet_heatmap * 0.4 + img
    superimposed_img = keras.preprocessing.image.array_to_img(superimposed_img)

    return superimposed_img



import unittest
import numpy as np
from dsbox.ml.markov.hmmlearn import hmm
from tests.ml.markov.hmmlearn import normalized, log_likelihood_increasing


class TestMultinomialAgainstWikipedia(unittest.TestCase):
    """Using examples from Wikipedia

    - http://en.wikipedia.org/wiki/Hidden_Markov_model
    - http://en.wikipedia.org/wiki/Viterbi_algorithm
    """

    def setup_method(self):
        n_components = 2  # ['Rainy', 'Sunny']
        n_features = 3  # ['walk', 'shop', 'clean']
        self.h = hmm.MultinomialHMM(n_components)
        self.h.n_features = n_features
        self.h.startprob_ = np.array([0.6, 0.4])
        self.h.transmat_ = np.array([[0.7, 0.3], [0.4, 0.6]])
        self.h.emissionprob_ = np.array([[0.1, 0.4, 0.5],
                                         [0.6, 0.3, 0.1]])

    def test_decode_viterbi(self):
        self.setup_method()
        # From http://en.wikipedia.org/wiki/Viterbi_algorithm:
        # "This reveals that the observations ['walk', 'shop', 'clean']
        #  were most likely generated by states ['Sunny', 'Rainy', 'Rainy'],
        #  with probability 0.01344."
        X = [[0], [1], [2]]
        logprob, state_sequence = self.h.decode(X, algorithm="viterbi")
        self.assertEqual(np.round(np.exp(logprob), 5), 0.01344)
        self.assertTrue(np.allclose(state_sequence, [1, 0, 0]))

    def test_decode_map(self):
        self.setup_method()
        X = [[0], [1], [2]]
        _logprob, state_sequence = self.h.decode(X, algorithm="map")
        self.assertTrue(np.allclose(state_sequence, [1, 0, 0]))

    def test_predict(self):
        self.setup_method()
        X = [[0], [1], [2]]
        state_sequence = self.h.predict(X)
        posteriors = self.h.predict_proba(X)
        self.assertTrue(np.allclose(state_sequence, [1, 0, 0]))
        self.assertTrue(np.allclose(posteriors, [
            [0.23170303, 0.76829697],
            [0.62406281, 0.37593719],
            [0.86397706, 0.13602294],
        ]))


class TestMultinomailHMM(unittest.TestCase):
    def setup_method(self):
        self.n_components = 2
        self.n_features = 3
        self.h = hmm.MultinomialHMM(self.n_components)
        self.h.startprob_ = np.array([0.6, 0.4])
        self.h.transmat_ = np.array([[0.7, 0.3], [0.4, 0.6]])
        self.h.emissionprob_ = np.array([[0.1, 0.4, 0.5], [0.6, 0.3, 0.1]])

    def test_attributes(self):
        self.setup_method()
        with self.assertRaises(ValueError):
            self.h.emissionprob_ = []
            self.h._check()

        with self.assertRaises(ValueError):
            self.h.emissionprob_ = np.zeros((self.n_components - 2,
                                             self.n_features))
            self.h._check()

    def test_score_samples(self):
        self.setup_method()
        idx = np.repeat(np.arange(self.n_components), 10)
        n_samples = len(idx)
        X = np.random.randint(self.n_features, size=(n_samples, 1))

        ll, posteriors = self.h.score_samples(X)
        self.assertEqual(posteriors.shape, (n_samples, self.n_components))
        self.assertTrue(np.allclose(posteriors.sum(axis=1), np.ones(n_samples)))

    def test_sample(self, n_samples=1000):
        self.setup_method()
        X, state_sequence = self.h.sample(n_samples)
        self.assertEqual(X.ndim, 2)
        self.assertTrue(len(X) == len(state_sequence) == n_samples)
        self.assertEqual(len(np.unique(X)), self.n_features)

    def test_fit(self, params='ste', n_iter=5):
        self.setup_method()
        h = self.h
        h.params = params

        lengths = np.array([10] * 10)
        X, _state_sequence = h.sample(lengths.sum())

        # Mess up the parameters and see if we can re-learn them.
        h.startprob_ = normalized(np.random.random(self.n_components))
        h.transmat_ = normalized(
            np.random.random((self.n_components, self.n_components)),
            axis=1)
        h.emissionprob_ = normalized(
            np.random.random((self.n_components, self.n_features)),
            axis=1)

        self.assertTrue(log_likelihood_increasing(h, X, lengths, n_iter))

    def test_fit_emissionprob(self):
        self.setup_method()
        self.test_fit('e')

    def test_fit_with_init(self, params='ste', n_iter=5):
        self.setup_method()
        lengths = [10] * 10
        X, _state_sequence = self.h.sample(sum(lengths))

        # use init_function to initialize paramerters
        h = hmm.MultinomialHMM(self.n_components, params=params,
                               init_params=params)
        h._init(X, lengths=lengths)

        self.assertTrue(log_likelihood_increasing(h, X, lengths, n_iter))



from functools import partial
from warnings import warn

import numpy as np

from ._utils import check_random_state

__credits__ = "Lars Buitinck"


def bio_f_score(y_true, y_pred):
    """F-score for BIO-tagging scheme, as used by CoNLL.

    This F-score variant is used for evaluating named-entity recognition and
    related problems, where the goal is to predict segments of interest within
    sequences and mark these as a "B" (begin) tag followed by zero or more "I"
    (inside) tags. A true positive is then defined as a BI* segment in both
    y_true and y_pred, with false positives and false negatives defined
    similarly.

    Support for tags schemes with classes (e.g. "B-NP") are limited: reported
    scores may be too high for inconsistent labelings.

    Parameters
    ----------
    y_true : array-like of strings, shape (n_samples,)
        Ground truth labeling.

    y_pred : array-like of strings, shape (n_samples,)
        Sequence classifier's predictions.

    Returns
    -------
    f : float
        F-score.
    """

    if len(y_true) != len(y_pred):
        msg = "Sequences not of the same length ({} != {})."""
        raise ValueError(msg.format(len(y_true), len(y_pred)))

    y_true = np.asarray(y_true)
    y_pred = np.asarray(y_pred)

    is_b = partial(np.char.startswith, prefix="B")

    where = np.where
    t_starts = where(is_b(y_true))[0]
    p_starts = where(is_b(y_pred))[0]

    # These lengths are off-by-one because we skip the "B", but that's ok.
    # http://stackoverflow.com/q/17929499/166749
    t_lengths = np.diff(where(is_b(np.r_[y_true[y_true != 'O'], ['B']]))[0])
    p_lengths = np.diff(where(is_b(np.r_[y_pred[y_pred != 'O'], ['B']]))[0])

    t_segments = set(zip(t_starts, t_lengths, y_true[t_starts]))
    p_segments = set(zip(p_starts, p_lengths, y_pred[p_starts]))

    # tp = len(t_segments & p_segments)
    # fn = len(t_segments - p_segments)
    # fp = len(p_segments - t_segments)
    tp = sum(x in t_segments for x in p_segments)
    fn = sum(x not in p_segments for x in t_segments)
    fp = sum(x not in t_segments for x in p_segments)

    if tp == 0:
        # special-cased like this in CoNLL evaluation
        return 0.

    precision = tp / float(tp + fp)
    recall = tp / float(tp + fn)

    return 2. * precision * recall / (precision + recall)


def whole_sequence_accuracy(y_true, y_pred, lengths):
    """Average accuracy measured on whole sequences.

    Returns the fraction of sequences in y_true that occur in y_pred without a
    single error.
    """
    lengths = np.asarray(lengths)
    end = np.cumsum(lengths)
    start = end - lengths
    bounds = np.vstack([start, end]).T

    errors = sum(1. for i, j in bounds
                 if np.any(y_true[i:j] != y_pred[i:j]))
    return 1 - errors / len(lengths)


class SequenceKFold(object):
    """Sequence-aware (repeated) k-fold CV splitter.

    Uses a greedy heuristic to partition input sequences into sets with roughly
    equal numbers of samples, while keeping the sequences intact.

    Parameters
    ----------
    lengths : array-like of integers, shape (n_samples,)
        Lengths of sequences, in the order in which they appear in the dataset.

    n_folds : int, optional
        Number of folds.

    n_iter : int, optional
        Number of iterations of repeated k-fold splitting. The default value
        is 1, meaning a single k-fold split; values >1 give repeated k-fold
        with shuffling (see below).

    shuffle : boolean, optional
        Whether to shuffle sequences.

    random_state : {np.random.RandomState, integer}, optional
        Random state/random seed for shuffling.

    yield_lengths : boolean, optional
        Whether to yield lengths in addition to indices/masks for both
        training and test sets.

    Returns
    -------
    folds : iterable
        A generator yielding (train_indices, test_indices) pairs when
        yield_lengths is false, or tuples
        (train_indices, train_lengths, test_indices, test_lengths)
        when yield_lengths is true.
    """

    def __init__(self, lengths, n_folds=3, n_iter=1, shuffle=False,
                 random_state=None, yield_lengths=True):
        if n_iter > 1 and not shuffle:
            warn("n_iter > 1 makes little sense without shuffling!")
        self.lengths = lengths
        self.n_folds = n_folds
        self.n_iter = n_iter
        self.random_state = random_state
        self.shuffle = shuffle
        self.yield_lengths = yield_lengths

    def __iter__(self):
        rng = check_random_state(self.random_state)
        lengths = np.asarray(self.lengths, dtype=np.int32)
        starts = np.cumsum(lengths) - lengths
        n_samples = np.sum(lengths)

        seq_ind = np.arange(len(lengths))

        for _ in range(self.n_iter):
            if self.shuffle:
                rng.shuffle(seq_ind)

            folds = [[] for _ in range(self.n_folds)]
            samples_per_fold = np.zeros(self.n_folds, dtype=int)

            # Greedy strategy: always append to the currently smallest fold
            for i in seq_ind:
                seq = (i, starts[i], starts[i] + lengths[i])
                fold_idx = np.argmin(samples_per_fold)
                folds[fold_idx].append(seq)
                samples_per_fold[fold_idx] += lengths[i]

            for f in folds:
                test = np.zeros(n_samples, dtype=bool)
                lengths_test_mask = np.zeros(len(lengths), dtype=bool)
                for i, start, end in f:
                    test[start:end] = True
                    lengths_test_mask[i] = True

                train = ~test
                train = np.where(train)[0]
                test = np.where(test)[0]
                lengths_test = lengths[lengths_test_mask]
                lengths_train = lengths[~lengths_test_mask]

                if self.yield_lengths:
                    yield train, lengths_train, test, lengths_test
                else:
                    yield train, test

    def __len__(self):
        return self.n_folds * self.n_iter



import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from dsbox.dbconnection.plasma_connector import PlasmaConnector
from tests.config import socket_name


class TestPlasmaConnector(unittest.TestCase):
    def test_put_and_get_dataframe_into_plasma_store(self):
        # given
        d = {'one': pd.Series([1., 2., 3.], index=['a', 'b', 'c']),
             'two': pd.Series([1., 2., 3., 4.], index=['a', 'b', 'c', 'd'])}
        df = pd.DataFrame(d)

        plasma_store = PlasmaConnector(socket_name)

        # when
        df_id = plasma_store.put_dataframe(df)
        df_from_store = plasma_store.get_dataframe(df_id)

        # then
        assert_frame_equal(df, df_from_store)

        plasma_store.client.disconnect()



import os
from google.cloud import bigquery

from dsbox.operators.data_unit import DataInputUnit, DataOutputUnit
from dsbox.utils.bq_client import get_bq_client, get_bq_storage_client


class JobError(Exception):
    pass


class DataInputBigQueryUnit(DataInputUnit):
    """
    Data unit allowing to load data from a SQL BigQuery query to a Pandas dataframe.

    Parameters
    ----------
    query: str
        BigQuery SQL query
    path_json_key: str, optional
        path to JSON GCP credentials file
    """

    def __init__(self, query, path_json_key=None):
        self.query = query
        self.path_json_key = path_json_key

    def read_data(self):
        if self.path_json_key is not None:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.path_json_key
        bqclient = get_bq_client()
        dataframe = bqclient.query(self.query).to_dataframe(bqstorage_client=get_bq_storage_client())
        return dataframe

    def __str__(self):
        return self.query


class DataOutputBigQueryUnit(DataOutputUnit):
    """
    Data unit allowing to write data from a Pandas dataframe to a BigQuery table.

    Parameters
    ----------
    table_id: str
        target BigQuery table using the format: dataset_name.table_name
    path_json_key: str, optional
        path to JSON GCP credentials file
    drop_table: bool, optional (default=True)
        replacing table if set to True (default)
    """
    def __init__(self, table_id, path_json_key=None, drop_table=True):
        self.table_id = table_id
        self.path_json_key = path_json_key
        self.drop_table = drop_table

    def write_data(self, dataframe):
        if self.path_json_key is not None:
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self.path_json_key
        if len(dataframe) > 0:
            str_cols = []
            for col in dataframe.columns:
                if dataframe[col].dtype == object:
                    print('col str type {}'.format(col))
                    str_cols.append(bigquery.SchemaField(col, "STRING"))

            job_config = bigquery.LoadJobConfig(schema=str_cols)
            if self.drop_table:
                job_config.write_disposition = "WRITE_TRUNCATE"

            bqclient = get_bq_client()
            try:
                job = bqclient.load_table_from_dataframe(dataframe, self.table_id, job_config=job_config)
                result = job.result()
                if result.state != 'DONE':
                    raise JobError()
            except JobError:
                print('BQ job error')
                print('Job state: {}'.format(result.state))
                print('Job errors: {}'.format(result.errors))
        else:
            print("Empty dataframe, no data to write to BQ table")

    def __str__(self):
        return self.table_id



import sys

from dsbox.kube.data_operations import Dataoperations

"""
Main program that should be used by a KubeDataOperator by passing it as 'arguments' 
parameters to execute operation on the data during the Pod execution.
"""

if __name__ == "__main__":
    assert len(sys.argv) > 1

    operation_name = sys.argv[1]
    path = sys.argv[2]
    operations_path = sys.argv[3]

    project_data_ops = Dataoperations(path=path)

    project_data_ops.load_datasets(operations_path)

    project_data_ops.run(operation_name)



from datetime import datetime
import os

from airflow import DAG

from dsbox.examples.sales_prediction_usecase.ml.feature_engineering import concat_train_test, resample_fillna, \
    create_simple_features, create_timeseries_rolling_features, create_timeseries_shift_features, \
    create_timeseries_diff_shift_features, merge_time_features, extra_features, merge_final_features
from dsbox.examples.sales_prediction_usecase.ml.modeling import train_model, metrics_model, predict_model
from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputFileUnit, DataInputMultiFileUnit

"""
This usecase is taken from the Kaggle Rossmann challenge.

"""

project_path = os.getenv('PROJECT_PATH')

dag = DAG(dag_id='Sales_Prediction', description='Sales Prediction Example',
          schedule_interval='0 20 * * *', start_date=datetime(2017, 3, 20), catchup=False)

"""
Concat train and test data (easier for feature engineering when facing a timeseries problem.
"""
input_parquet_files_unit = DataInputMultiFileUnit([project_path + 'datasets/input/train.parquet',
                                                   project_path + 'datasets/input/test.parquet'],
                                                  pandas_read_function_name='read_parquet')

output_parquet_concat_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_raw.parquet',
                                                pandas_write_function_name='to_parquet')

task_concat_train_files = DataOperator(operation_function=concat_train_test,
                                       input_unit=input_parquet_files_unit,
                                       output_unit=output_parquet_concat_unit,
                                       dag=dag, task_id='Concat_train_test_data_source_files')

"""
Resampling time data
"""
input_raw_data_unit = DataInputFileUnit(output_parquet_concat_unit.output_path,
                                        pandas_read_function_name='read_parquet')

output_cleaned_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_clean.parquet',
                                              pandas_write_function_name='to_parquet')

task_fill_missing_values = DataOperator(operation_function=resample_fillna,
                                        input_unit=input_raw_data_unit,
                                        output_unit=output_cleaned_data_unit,
                                        dag=dag, task_id='Resample_and_fill_NA_values')

task_concat_train_files.set_downstream(task_fill_missing_values)

"""
Simple feature engineering 
"""

input_cleaned_data_unit = DataInputFileUnit(output_cleaned_data_unit.output_path,
                                            pandas_read_function_name='read_parquet')

output_simple_fe_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_simple_fe.parquet',
                                                pandas_write_function_name='to_parquet')

task_simple_fe = DataOperator(operation_function=create_simple_features,
                              input_unit=input_cleaned_data_unit,
                              output_unit=output_simple_fe_data_unit,
                              dag=dag, task_id='Simple_feature_engineering')

task_fill_missing_values.set_downstream(task_simple_fe)

"""
Time feature engineering
"""

operations = ['mean', 'std', 'median', 'min', 'max']
shift_ranges = [42, 49, 56, 63] + list(range(70, 364, 28)) + [364]
rolling_windows = [7, 28]

input_cleaned_sub_data_unit = DataInputFileUnit(output_simple_fe_data_unit.output_path,
                                                pandas_read_function_name='read_parquet',
                                                columns=['Date', 'Store', 'Sales',
                                                         'Customers', 'SalePerCustomer'])
rolling_tasks = []

for operation in operations:
    output_rolling_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_' + operation + '.parquet',
                                                  pandas_write_function_name='to_parquet')

    task_rolling = DataOperator(operation_function=create_timeseries_rolling_features,
                                params={'features': ['Sales', 'Customers', 'SalePerCustomer'],
                                        'operation': operation,
                                        'shift_ranges': shift_ranges,
                                        'rolling_windows': rolling_windows},
                                input_unit=input_cleaned_sub_data_unit,
                                output_unit=output_rolling_data_unit,
                                dag=dag, task_id='Rolling_' + operation + '_feature_engineering')

    rolling_tasks.append(task_rolling)
    task_simple_fe.set_downstream(task_rolling)

output_shift_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_shift.parquet',
                                            pandas_write_function_name='to_parquet')

task_shift = DataOperator(operation_function=create_timeseries_shift_features,
                          params={'features': ['Sales', 'Customers', 'SalePerCustomer'],
                                  'shift_ranges': shift_ranges},
                          input_unit=input_cleaned_sub_data_unit,
                          output_unit=output_shift_data_unit,
                          dag=dag, task_id='Time_shift_feature_engineering')

task_simple_fe.set_downstream(task_shift)

input_shift_data_unit = DataInputFileUnit(output_shift_data_unit.output_path,
                                          pandas_read_function_name='read_parquet')
output_diff_shift_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_diff_shift.parquet',
                                                 pandas_write_function_name='to_parquet')

task_diff_shift = DataOperator(operation_function=create_timeseries_diff_shift_features,
                               params={'features': ['Sales', 'Customers', 'SalePerCustomer'],
                                       'shift_ranges': shift_ranges},
                               input_unit=input_shift_data_unit,
                               output_unit=output_diff_shift_data_unit,
                               dag=dag, task_id='Time_diff_shift_feature_engineering')

task_shift.set_downstream(task_diff_shift)

paths_to_merge = [output_simple_fe_data_unit.output_path,
                  output_shift_data_unit.output_path,
                  output_diff_shift_data_unit.output_path]
for operation in operations:
    paths_to_merge.append(project_path + 'datasets/temp/X_' + operation + '.parquet')

input_time_files_unit = DataInputMultiFileUnit(paths_to_merge, pandas_read_function_name='read_parquet')

output_time_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_time.parquet',
                                           pandas_write_function_name='to_parquet')

task_merge_time_features = DataOperator(operation_function=merge_time_features,
                                        params={'features': ['Sales', 'Customers', 'SalePerCustomer']},
                                        input_unit=input_time_files_unit,
                                        output_unit=output_time_data_unit,
                                        dag=dag, task_id='Time_merge_feature_engineering')

task_diff_shift.set_downstream(task_merge_time_features)
for rolling_task in rolling_tasks:
    rolling_task.set_downstream(task_merge_time_features)

"""
Extra data feature engineering
"""
input_extra_data_unit = DataInputFileUnit(project_path + 'datasets/input/store.csv')
output_extra_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_extra.parquet',
                                            pandas_write_function_name='to_parquet')

task_extra_features = DataOperator(operation_function=extra_features,
                                   input_unit=input_extra_data_unit,
                                   output_unit=output_extra_data_unit,
                                   dag=dag, task_id='Extra_data_feature_engineering')

"""
Final features merge
"""
input_merge_files_unit = DataInputMultiFileUnit([project_path + 'datasets/temp/X_time.parquet',
                                                 project_path + 'datasets/temp/X_extra.parquet'],
                                                pandas_read_function_name='read_parquet')
output_merge_data_unit = DataOutputFileUnit(project_path + 'datasets/temp/X_final.parquet',
                                            pandas_write_function_name='to_parquet')

task_final_features = DataOperator(operation_function=merge_final_features,
                                   input_unit=input_merge_files_unit,
                                   output_unit=output_merge_data_unit,
                                   dag=dag, task_id='Final_data_merge')

task_extra_features.set_downstream(task_final_features)
task_merge_time_features.set_downstream(task_final_features)

"""
Train model
"""
input_final_data_unit = DataInputFileUnit(output_merge_data_unit.output_path,
                                          pandas_read_function_name='read_parquet')

features_to_exclude = ['Datetime', 'Date', 'Sales', 'Customers', 'SalePerCustomer', 'Id', 'to_predict', 'Open',
                       'MissingValues']

task_train_model = DataOperator(operation_function=train_model,
                                params={'features_to_exclude': features_to_exclude,
                                        'model_path': project_path + 'models/'},
                                input_unit=input_final_data_unit,
                                dag=dag, task_id='Train_model')

task_final_features.set_downstream(task_train_model)

"""
Metrics and prediction
"""

task_metrics = DataOperator(operation_function=metrics_model,
                            params={'features_to_exclude': features_to_exclude},
                            input_unit=input_final_data_unit,
                            dag=dag, task_id='Metrics')

task_train_model.set_downstream(task_metrics)

output_predict_data_unit = DataOutputFileUnit(project_path + 'datasets/output/X_predict.csv', index=False)

task_predict = DataOperator(operation_function=predict_model,
                            params={'features_to_exclude': features_to_exclude,
                                    'model_path': project_path + 'models/'},
                            input_unit=input_final_data_unit,
                            output_unit=output_predict_data_unit,
                            dag=dag, task_id='Prediction')

task_train_model.set_downstream(task_predict)



import numpy as np

__author__ = "Samuel Rochette"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def compute_prediction_interval(model, data_eval, percentile=95):
    """
    A function that compute the prediction interval based on the predictions of each trees of a random forest.

        Parameters
        ----------
        model: a fitted RandomForestRegressor.

        data_eval: A panda or numpy matrix of data of the validation set.

        percentile: Percentile in the interval. Default value 95.

        Returns
        -------
        Two arrays : lower_bound_prediction_list and upper_bound_prediction_list

        References
        ----------
        [1]  Quantile Regression Forests, Nicolai Meinshausen. 2006
        [2]  https://blog.datadive.net/prediction-intervals-for-random-forests/
    """

    decision_tree_list = model.estimators_
    lower_bound = (100 - percentile) / 2.
    upper_bound = 100 - lower_bound

    # this is the matrix of each predictions for each trees
    decision_tree_prediction_matrix = [_compute_prediction_of_each_tree(decision_tree_list, data_eval[i]) for i in
                                       range(len(data_eval))]

    lower_bound_prediction_list = [np.percentile(prediction_list, lower_bound) for prediction_list in
                                   decision_tree_prediction_matrix]
    upper_bound_prediction_list = [np.percentile(prediction_list, upper_bound) for prediction_list in
                                   decision_tree_prediction_matrix]

    return lower_bound_prediction_list, upper_bound_prediction_list


def _compute_prediction_of_each_tree(decision_tree_list, row):
    # reshaping is mandatory to avoid scikit warning
    reshaped_row = row.reshape(1, -1)
    return [decision_tree.predict(reshaped_row)[0] for decision_tree in decision_tree_list]



import yaml
from importlib import import_module

from dsbox.operators.data_executor import DataExecutor
from dsbox.utils import format_dict_path_items


class Dataoperations:
    """
    Load data operations meta-data (YAML format).

    Ex:

    Join_train_data_source_files:
      operation_function:
        module: tree_disease.ml.feature_engineering
        name: join_dataframes
      input_unit:
        type: DataInputMultiFileUnit
        input_path_list:
          - '{}/X_tree_egc_t1.csv'
          - '{}/X_geoloc_egc_t1.csv'
          - '{}/Y_tree_egc_t1.csv'
        pandas_read_function_name: read_csv
        sep: ';'
      output_unit:
        type: DataOutputFileUnit
        output_path: '{}/X_train_raw.parquet'
        pandas_write_function_name: to_parquet

    """

    def __init__(self, path=None, data_unit_module='dsbox.operators.data_unit',
                 data_input_unit_module=None,
                 data_output_unit_module=None):
        self.path = path
        self.parsed_datasets_file = None
        self.data_unit_module = data_unit_module
        self.data_input_unit_module = data_input_unit_module
        self.data_output_unit_module = data_output_unit_module
        if self.data_input_unit_module is None:
            self.data_input_unit_module = self.data_unit_module
        if self.data_output_unit_module is None:
            self.data_output_unit_module = self.data_unit_module

    def load_datasets(self, datasets_file_path):
        """
        Add a datasets operations catalog.

        Parameters
        ----------
        datasets_file_path: str
            datasets catalog operations path

        """
        datasets_file = open(datasets_file_path)
        self.parsed_datasets_file = yaml.load(datasets_file, Loader=yaml.FullLoader)

    def run(self, operation_name):
        """
        Run operation defined in data catalog operation.

        Parameters
        ----------
        operation_name: str
            operation name to run

        """
        if self.parsed_datasets_file is None:
            raise AttributeError('Call load_datasets method before running operation. '
                                 'Datasets catalog operations is not set.')

        operation_structure = self.parsed_datasets_file[operation_name]
        operation_structure = format_dict_path_items(operation_structure, self.path)

        operation_infos = operation_structure['operation_function']
        operation = getattr(import_module(operation_infos['module']), operation_infos['name'])
        op_kwargs = dict()
        if 'kwargs' in operation_infos:
            op_kwargs = operation_infos['kwargs']

        input_unit = None
        output_unit = None

        if 'input_unit' in operation_structure:
            input_unit_structure = operation_structure['input_unit']
            DataInputUnitClass = getattr(import_module(self.data_input_unit_module), input_unit_structure['type'])
            parameters = input_unit_structure.copy()
            parameters.pop('type')
            input_unit = DataInputUnitClass(**parameters)
            print('Input unit: {}'.format(input_unit))

        if 'output_unit' in operation_structure:
            output_unit_structure = operation_structure['output_unit']
            DataOutputUnitClass = getattr(import_module(self.data_output_unit_module), output_unit_structure['type'])
            parameters = output_unit_structure.copy()
            parameters.pop('type')
            output_unit = DataOutputUnitClass(**parameters)
            print('Output unit: {}'.format(output_unit))

        task = DataExecutor(operation, input_unit=input_unit, output_unit=output_unit, **op_kwargs)
        task.execute()



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from sklearn.metrics import roc_curve, auc

__author__ = "Aurélien Massiot"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def plot_confusion_matrix(confusion_matrix, classes_list, normalize=True, figsize=(10, 7), fontsize=14, cmap="Blues"):
    """
    Display a pretty confusion matrix.

    Parameters
    ----------
    confusion_matrix : array-like

    classes_list : list,
        classes list of the confusion matrix

    normalize : boolean,
        normalize confusion matrix

    figsize : tuple, optional (default=(10,7))
        set the figure size

    fontsize : int, optional (default=14)
        set the font size

    cmap : str, optional (default="Blues")
        set the colormap

    Returns
    -------
    Confusion matrix figure


    Examples
    --------
    >>> from dsbox.ml.visualization.metrics import plot_confusion_matrix
    >>> array = [[ 8458,   227,  1730], \
             [ 1073, 37590,  1613], \
             [ 2390,  1159, 17540]]
    >>> classes_list = ["A", "B", "C"]
    >>> plot_confusion_matrix(array, classes_list)
   """
    confusion_matrix = np.array(confusion_matrix)

    fig, ax = plt.subplots(figsize=figsize)

    if normalize:
        normalized_cm = np.array(confusion_matrix).astype('float') / np.array(confusion_matrix).sum(axis=1)[:,
                                                                     np.newaxis]
        df_cm = pd.DataFrame(
            normalized_cm, index=classes_list, columns=classes_list,
        )
        plt.matshow(df_cm, fignum=0, cmap=cmap)
    else:
        df_cm = pd.DataFrame(
            confusion_matrix, index=classes_list, columns=classes_list,
        )
        plt.matshow(df_cm, fignum=0, cmap=cmap)
    ax.set_xticks(np.arange(len(classes_list)))
    ax.set_yticks(np.arange(len(classes_list)))
    ax.set_xticklabels(classes_list)
    ax.set_yticklabels(classes_list)
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

    for i in range(len(classes_list)):
        for j in range(len(classes_list)):
            ax.text(j, i, confusion_matrix[i, j], ha="center", va="center", color="grey", fontsize=fontsize)

    plt.ylabel('True labels')
    plt.xlabel('Predicted labels')
    plt.show()


def plot_roc_curve(y_test, y_pred_probas, proba_step=None):
    """
    Plot ROC curve with probabilities thresholds.
    
    Parameters
    ----------
    y_test : array-like
        true labels
    
    y_pred_probas : array-like
        predicted labels
    
    proba_step : int (optional) (default=None)
        if set, give the step for each probability display. If None, nothing is displayed.

    Examples
    --------
    
    >>> from dsbox.ml.visualization.metrics import plot_roc_curve
    >>> from sklearn import datasets
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.ensemble import RandomForestClassifier
    
    >>> X, y = datasets.make_moons(noise=0.3, random_state=0)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=.5, random_state=0)
    
    >>> clf = RandomForestClassifier(n_estimators=10, random_state=42)
    >>> _ = clf.fit(X_train, y_train)
    >>> y_pred_probas = clf.predict_proba(X_test)
    
    >>> plot_roc_curve(y_test, y_pred_probas, proba_step=2)

    """
    fpr, tpr, thresholds = roc_curve(y_test, y_pred_probas[:, 1])
    auc_score = auc(fpr, tpr)

    plt.figure()
    lw = 1
    plt.plot(fpr, tpr, color='darkorange', lw=lw, marker='.')
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    if proba_step is not None:
        i = 0
        for x, y, txt in zip(fpr, tpr, thresholds):
            if i % proba_step == 0:
                plt.annotate(np.round(txt, 2), (x, y - 0.04), color='darkgray', fontsize=8)
            i += 1
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.05])
    plt.xlabel('False Positive Rate')
    plt.ylabel('True Positive Rate')
    plt.title('Receiver operating characteristic (ROC) - AUC score: {}'.format(str(np.round(auc_score,3))))
    plt.show()



from abc import ABC, abstractmethod
from importlib import import_module

import pandas as pd


class DataInputUnit(ABC):
    """
    Abstract class defining reading contract for all Data Unit with input data.
    """
    input_path = None

    @abstractmethod
    def read_data(self):
        pass

    def __str__(self):
        return str(self.input_path)


class DataOutputUnit(ABC):
    """
    Abstract class defining reading contract for all Data Unit with output data.
    """
    output_path = None

    @abstractmethod
    def write_data(self, dataframe):
        pass

    def __str__(self):
        return str(self.output_path)


class DataGlobalInputUnit(DataInputUnit):
    """
    Data unit allowing to read data via any API using a dataframe structure (Pandas, Dask, Vaex).

    Parameters
    ----------
    input_path: str
        file path to read
    api_module: str, default='pandas'
        module name used apply read function set in read_function parameter
    read_function_name: str, default='read_csv'
        set the function name used by backend API to read data
    kwargs: dict
        used by backend API to pass others parameters
    """

    def __init__(self, input_path, api_module='pandas', read_function_name='read_csv', **kwargs):
        self.input_path = input_path
        self.read_function_name = read_function_name
        self.api_module = import_module(api_module)
        self.kwargs_read = kwargs

    def read_data(self):
        dataframe = getattr(self.api_module, self.read_function_name)(self.input_path, **self.kwargs_read)
        return dataframe


class DataGlobalOutputFileUnit(DataOutputUnit):
    """
    Data unit allowing to write data via dataframe API.

    Parameters
    ----------
    output_path: str
        file path to write
    write_function_name: str, default='to_csv'
        set the function name used by backend API to write data
    kwargs: dict
        used by backend API to pass others parameters
    """

    def __init__(self, output_path, write_function_name='to_csv', **kwargs):
        self.output_path = output_path
        self.write_function_name = write_function_name
        self.kwargs_write = kwargs

    def write_data(self, dataframe):
        getattr(dataframe, self.write_function_name)(self.output_path, **self.kwargs_write)


class DataInputFileUnit(DataInputUnit):
    """
    Data unit allowing to read data via Pandas API.

    Parameters
    ----------
    input_path: str
        file path to read
    pandas_read_function_name: str, default='read_csv'
        set the function name used by Pandas to read data
    kwargs: dict
        used by Pandas API to pass others parameters
    """

    def __init__(self, input_path, pandas_read_function_name='read_csv', **kwargs):
        self.input_path = input_path
        self.pandas_read_function_name = pandas_read_function_name
        self.pandas_kwargs_read = kwargs

    def read_data(self):
        dataframe = getattr(pd, self.pandas_read_function_name)(self.input_path, **self.pandas_kwargs_read)
        return dataframe


class DataInputPathUnit(DataInputUnit):
    """
    Lazy version of the DataInputFileUnit class, passing only the path to read_data method,
    not the loaded dataframe.

    Parameters
    ----------
    input_path: str
        file path to read
    """

    def __init__(self, input_path):
        self.input_path = input_path

    def read_data(self):
        return self.input_path


class DataOutputFileUnit(DataGlobalOutputFileUnit):
    """
    Data unit allowing to write data via Pandas API. (kept for backwards compatibility)

    Parameters
    ----------
    output_path: str
        file path to write
    pandas_write_function_name: str, default='to_csv'
        set the function name used by Pandas to write data
    kwargs: dict
        used by Pandas API to pass others parameters
    """

    def __init__(self, output_path, pandas_write_function_name='to_csv', **kwargs):
        super(DataOutputFileUnit, self).__init__(output_path, pandas_write_function_name, **kwargs)

    def write_data(self, dataframe):
        getattr(dataframe, self.write_function_name)(self.output_path, **self.kwargs_write)


class DataInputPlasmaUnit(DataInputUnit):
    """
    Data unit allowing to read persisted Pandas dataframe in memory using Arrow Plasma store.

    Parameters
    ----------
    plasma_store: PlasmaConnector
        object used to connect to Plasma store
    object_id: int
        unique id used to read dataframe
    """

    def __init__(self, plasma_store, object_id):
        self.plasma_store = plasma_store
        self.object_id = object_id

    def read_data(self):
        return self.plasma_store.get_dataframe(self.object_id)

    def __str__(self):
        return "{}/{}".format(self.plasma_store, str(self.object_id))


class DataOutputPlasmaUnit(DataOutputUnit):
    """
    Data unit allowing to persist Pandas dataframe in memory using Arrow Plasma store.

    Parameters
    ----------
    plasma_store: PlasmaConnector
        object used to connect to Plasma store
    object_id: int
        unique id used to write dataframe
    overwrite: bool, default=True
        if True, don't verify if object id is still present, and overwrite data, else, an exception is raised
    """

    def __init__(self, plasma_store, object_id, overwrite=True):
        self.plasma_store = plasma_store
        self.object_id = object_id
        self.overwrite = overwrite

    def write_data(self, dataframe):
        self.plasma_store.put_dataframe(dataframe, self.object_id, overwrite=self.overwrite)

    def __str__(self):
        return "{}/{}".format(self.plasma_store, str(self.object_id))


class DataInputMultiFileUnit(DataInputUnit):
    """
    Data unit allowing to read several data sources at once via Pandas API. The read_data method returns
    a list of loaded dataframes.
    Note: data format has to be the same for all data sources.

    Parameters
    ----------
    input_path_list: list
        list of paths data to read
    pandas_read_function_name: str, default='read_csv'
        set the function name used by Pandas to read data
    kwargs: dict
        used by Pandas API to pass others parameters
    """

    def __init__(self, input_path_list, pandas_read_function_name='read_csv', **kwargs):
        self.input_path_list = input_path_list
        self.pandas_read_function_name = pandas_read_function_name
        self.pandas_kwargs_read = kwargs

    def read_data(self):
        dataframe_list = []
        for input_path in self.input_path_list:
            dataframe_list.append(getattr(pd, self.pandas_read_function_name)(input_path, **self.pandas_kwargs_read))
        return dataframe_list

    def __str__(self):
        return str(self.input_path_list)


class DataInputMultiPathUnit(DataInputUnit):
    """
    Lazy version of the DataInputMultiFileUnit class, passing only the path to read_data method,
    not the loaded dataframe.

    Parameters
    ----------
    input_path_list: list
        list of paths data to read
    """

    def __init__(self, input_path_list):
        self.input_path_list = input_path_list

    def read_data(self):
        return self.input_path_list

    def __str__(self):
        return str(self.input_path_list)



import unittest
import numpy as np

from dsbox.ml.ensemble.rrclassifier.randomrotation import random_rotation_matrix
from dsbox.ml.ensemble.rrclassifier import RRForestClassifier, RRExtraTreesClassifier
from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score


class TestRRClassifier(unittest.TestCase):

    def test_random_rotation(self):
        R = random_rotation_matrix(3)

        self.assertTrue(np.allclose(R.T, np.linalg.inv(R)))

        self.assertIsNotNone(np.linalg.det(R))

    def test_RRForestClassifier(self):
        np.random.seed(42)

        # given
        X = load_iris()
        X_train, X_test, y_train, y_test = train_test_split(X.data, X.target, test_size=0.33, stratify=X.target,
                                                            random_state=42)
        # when
        clf = RRForestClassifier(n_estimators=10, random_state=42)
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)

        # then
        expected_score = 0.9
        f_score = f1_score(y_test, y_pred, average='micro')
        self.assertGreater(f_score, expected_score)

    def test_RRExtraTreesClassifier(self):
        np.random.seed(42)

        # given
        X = load_iris()
        X_train, X_test, y_train, y_test = train_test_split(X.data, X.target, test_size=0.33, stratify=X.target,
                                                            random_state=42)
        # when
        clf = RRExtraTreesClassifier(n_estimators=10, random_state=42)
        clf.fit(X_train, y_train)
        y_pred = clf.predict(X_test)

        # then
        expected_score = 0.9
        f_score = f1_score(y_test, y_pred, average='micro')
        self.assertGreater(f_score, expected_score)



import numpy as np
from dsbox.ml.utils import check_estimator_predict_proba

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_X_y, check_is_fitted, check_array

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class BinaryCalibrator(BaseEstimator, ClassifierMixin):
    """ Probability calibration using Bayes Minimum Risk theory to calibrate prediction 
    with unbalanced binary target.
    
    It is known that undersampling one class modifies the priors of the training set and 
    consequently biases the posterior probabilities of a classifier. This class finds the
    correct classification threshold and how to adjust it after undersampling to keep
    unbiased posterior prediction probabilities.
    
    Remark: this method concerns the probabilities of prediction for the underrepresented class (p').
    Probabilities for the major class are just computed as 1 - p'.
            

        Parameters
        ----------
        base_estimator : BaseEstimator
            estimator to calibrate
            
        major_class_proportion: float, optional, default 1.0
            Proportion of the major class to consider for undersampling.
            If major_class_proportion == 1, then the balance will be 50/50. To undersample less, reduce 
            this parameter.
       
        References
        ----------
        .. [1]  Andrea Dal Pozzolo and al. Calibrating Probability with Undersampling for 
                Unbalanced Classification. IEEE Symposium Series on Computational Intelligence, 
                SSCI 2015, 159-166

        See also
        --------
        sklearn.calibration.CalibratedClassifierCV : Probability calibration with isotonic regression or sigmoid
        
    """

    def __init__(self, base_estimator, major_class_proportion=1.0):

        self.base_estimator = base_estimator

        check_estimator_predict_proba(self.base_estimator)

        self.major_class_value = 0
        self.minor_class_value = 1

        self.major_class_proportion = major_class_proportion

        self.attr_to_check = ["major_class_amount_",
                              "minor_class_amount_",
                              "beta_",
                              "tau_prime_",
                              "X_",
                              "y_",
                              "classes_"]

    def fit(self, X, y):
        """ Fit the calibrated model
        
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data.
            
        y : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        check_X_y(X, y)

        unique_y = set(y)

        if len(unique_y) != 2:
            raise ValueError('BinaryCalibrator can only fit binary class, y modalities = ' + str(len(unique_y)))
        if 0 not in unique_y or 1 not in unique_y:
            raise ValueError('Target class value must be 0 and 1.')

        self.classes_ = list(unique_y)

        # check which class is in majority
        np_y = np.array(y).astype('int')
        val_distrib = np.bincount(np_y)
        if val_distrib[0] < val_distrib[1]:
            self.major_class_value = 1
            self.minor_class_value = 0

        # get amount of elements for each class
        self.major_class_amount_ = val_distrib[self.major_class_value]
        self.minor_class_amount_ = val_distrib[self.minor_class_value]

        # compute beta: ratio of the prior probabilities
        self.beta_ = (self.minor_class_amount_ / float(len(y))) / (self.major_class_amount_ / float(len(y)))

        # shuffle major class indexes for undersampling

        self.X_ = X.copy()
        self.X_['y'] = y
        self.X_['y'] = self.X_['y'].astype('int')
        X_major_class_indexes = np.array(self.X_[self.X_['y'] == self.major_class_value].index)
        np.random.shuffle(X_major_class_indexes)

        # remove data from major class to balance the dataset
        amount_index_to_remove = int((1 - self.beta_) * (self.major_class_amount_ * self.major_class_proportion))
        index_to_remove = X_major_class_indexes[0:amount_index_to_remove]
        df_X_to_remove = self.X_.iloc[index_to_remove]
        self.X_ = self.X_.drop(df_X_to_remove.index)
        self.y_ = self.X_['y']

        # tau_prime: threshold that guarantees equal accuracy after the posterior probability correction
        self.tau_s_ = len(self.X_[self.X_['y'] == self.major_class_value]) / float(len(self.X_))
        self.tau_prime_ = (self.beta_ * self.tau_s_) / ((self.beta_ - 1) * self.tau_s_ + 1)

        del self.X_['y']

        self.base_estimator.fit(self.X_, self.y_)

        return self

    def predict_proba(self, X):
        """Posterior probabilities of classification
        This function returns posterior probabilities of classification
        according to beta parameter computed during fit
        
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The samples.
            
        Returns
        -------
        y_probas : array, shape (n_samples, n_classes)
            The predicted probas.
        """
        check_is_fitted(self, self.attr_to_check)

        check_array(X)

        y_probas = self.base_estimator.predict_proba(X)
        for i in range(0, len(y_probas)):
            yp = y_probas[i]
            yp[self.minor_class_value] = self._bias_corrected_probability(yp[self.minor_class_value])
            yp[self.major_class_value] = 1 - yp[self.minor_class_value]
            y_probas[i] = yp  # should be useless as yp is a reference

        return y_probas

    def predict(self, X):
        """Predict the target of new samples. Can be different from the
        prediction of the uncalibrated classifier.
        
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The samples.
            
        Returns
        -------
        C : array, shape (n_samples,)
            The predicted class.
        """

        y_pred_probas = self.predict_proba(X)
        y_pred = []
        for yp in y_pred_probas:
            if yp[1] >= 0.5:
                y_pred.append(1)
            else:
                y_pred.append(0)

        return y_pred

    def _bias_corrected_probability(self, p):

        return (self.beta_ * p) / ((self.beta_ * p) - p + 1)

    def bias_corrected_threshold(self):
        """
        Returns the threshold that guarantees equal accuracy after the posterior probability correction.
        
        Returns
        -------
        tau_prime_ : float
            tau' threshold
        """

        return self.tau_prime_



from airflow import DAG

from dsbox.examples.tree_disease_usecase.ml.feature_engineering import fillna_columns, category_to_numerical_features
from dsbox.operators.data_operator import DataOperator
from dsbox.operators.data_unit import DataInputFileUnit, DataOutputFileUnit


def feature_engineering_sub_dag(parent_dag_name, child_dag_name, model_path, input_file, output_file,
                                temp_files, start_date,
                                schedule_interval, mode='train'):


    dag = DAG('%s.%s' % (parent_dag_name, child_dag_name),
              schedule_interval=schedule_interval,
              start_date=start_date)

    task_fillna = DataOperator(operation_function=fillna_columns,
                               input_unit=DataInputFileUnit(input_file, pandas_read_function_name='read_parquet'),
                               output_unit=DataOutputFileUnit(temp_files[0], pandas_write_function_name='to_parquet'),
                               dag=dag, task_id='Fill_NA_values',
                               params={
                                   'simple_features': ['NOTEDIAGNOSTIC', 'PRIORITEDERENOUVELLEMENT',
                                                       'FREQUENTATIONCIBLE', 'RAISONDEPLANTATION',
                                                       'SOUS_CATEGORIE', 'STADEDEDEVELOPPEMENT',
                                                       'STADEDEVELOPPEMENTDIAG',
                                                       'TRAITEMENTCHENILLES', 'TRAVAUXPRECONISESDIAG', 'TROTTOIR',
                                                       'VARIETE', 'VIGUEUR', 'CODE_PARENT'],
                                   'model_path': model_path,
                                   'mode': mode}
                               )

    task_cat_to_num = DataOperator(operation_function=category_to_numerical_features,
                                   input_unit=DataInputFileUnit(temp_files[0],
                                                                pandas_read_function_name='read_parquet'),
                                   output_unit=DataOutputFileUnit(output_file, pandas_write_function_name='to_parquet'),
                                   dag=dag, task_id='Categorical_features_to_numeric',
                                   params={'features': ['GENRE_BOTA', 'ESPECE',
                                                        'FREQUENTATIONCIBLE', 'RAISONDEPLANTATION',
                                                        'SOUS_CATEGORIE', 'STADEDEDEVELOPPEMENT',
                                                        'STADEDEVELOPPEMENTDIAG',
                                                        'TRAITEMENTCHENILLES', 'TRAVAUXPRECONISESDIAG', 'TROTTOIR',
                                                        'VARIETE', 'VIGUEUR', 'CODE_PARENT'],
                                           'model_path': model_path,
                                           'mode': mode}
                                   )

    task_fillna.set_downstream(task_cat_to_num)

    return dag



import numpy as np
from sklearn.ensemble._forest import ForestClassifier
from sklearn.tree import ExtraTreeClassifier
from dsbox.ml.ensemble.rrclassifier.randomrotation import random_rotation_matrix

from sklearn.ensemble._base import _partition_estimators
from joblib import Parallel, delayed
from scipy.stats.mstats_basic import mquantiles

__credits__ = "https://github.com/tmadl/sklearn-random-rotation-ensembles"


def _parallel_helper(obj, methodname, *args, **kwargs):
    """Private helper to workaround Python 2 pickle limitations"""
    return getattr(obj, methodname)(*args, **kwargs)


class RRExtraTreeClassifier(ExtraTreeClassifier):
    def __init__(self,
                 criterion="gini",
                 splitter="random",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 random_state=None,
                 max_leaf_nodes=None,
                 class_weight=None):
        super(RRExtraTreeClassifier, self).__init__(
            criterion=criterion,
            splitter=splitter,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            class_weight=class_weight,
            random_state=random_state)

    def rotate(self, X):
        if not hasattr(self, 'rotation_matrix'):
            raise Exception('The estimator has not been fitted')

        return np.dot(X, self.rotation_matrix)

    def _fit_rotation_matrix(self, X):
        # self.rotation_matrix = np.eye(X.shape[1]).astype(np.float32)
        self.rotation_matrix = random_rotation_matrix(X.shape[1])

    def fit(self, X, y, sample_weight=None, check_input=True):
        self._fit_rotation_matrix(X)
        super(RRExtraTreeClassifier, self).fit(self.rotate(X), y,
                                               sample_weight, check_input)

    def predict_proba(self, X, check_input=True):
        return super(RRExtraTreeClassifier, self).predict_proba(self.rotate(X),
                                                                check_input)

    def predict(self, X, check_input=True):
        return super(RRExtraTreeClassifier, self).predict(self.rotate(X),
                                                          check_input)

    def apply(self, X, check_input=True):
        return super(RRExtraTreeClassifier, self).apply(self.rotate(X),
                                                        check_input)

    def decision_path(self, X, check_input=True):
        return super(RRExtraTreeClassifier, self).decision_path(self.rotate(X),
                                                                check_input)


class RRExtraTreesClassifier(ForestClassifier):
    """

        Parameters
        ----------
        see sklearn documentation
        http://scikit-learn.org/stable/modules/generated/sklearn.ensemble.ExtraTreesClassifier.html
    
        Examples
        --------
        
        >>> from sklearn.datasets import load_iris
        >>> from sklearn.model_selection import train_test_split
        >>> from dsbox.ml.ensemble.rrclassifier import RRExtraTreesClassifier
        
        >>> import numpy as np
        >>> np.random.seed(42)
        
        >>> X = load_iris()
        >>> X_train, X_test, y_train, y_test = train_test_split(X.data, X.target, test_size=0.33, stratify=X.target, random_state=42)
        
        >>> clf = RRExtraTreesClassifier(random_state=42)
        >>> _ = clf.fit(X_train, y_train)
        >>> clf.predict(X_test)
        array([1, 1, 0, 1, 2, 1, 1, 0, 1, 1, 0, 0, 0, 0, 0, 2, 2, 1, 2, 1, 2, 1,
               0, 2, 0, 2, 2, 0, 0, 2, 2, 2, 0, 1, 0, 0, 2, 1, 2, 1, 1, 1, 0, 0,
               2, 1, 2, 1, 1, 2])
        
        References
        ----------
        
        original package: https://github.com/tmadl/sklearn-random-rotation-ensembles
        
        Implementation of the Algorithm: J. J. Rodriguez, L. I. Kuncheva and C. J. Alonso, 
        "Rotation Forest: A New Classifier Ensemble Method," in IEEE Transactions on Pattern 
        Analysis and Machine Intelligence, vol. 28, no. 10, pp. 1619-1630, Oct. 2006.
        doi: 10.1109/TPAMI.2006.211
        
    """

    def __init__(self,
                 n_estimators=10,
                 criterion="gini",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 bootstrap=False,
                 oob_score=False,
                 n_jobs=1,
                 random_state=None,
                 verbose=0,
                 warm_start=False,
                 class_weight=None,
                 scaling=True):

        super(RRExtraTreesClassifier, self).__init__(
            base_estimator=RRExtraTreeClassifier(),
            n_estimators=n_estimators,
            estimator_params=("criterion", "max_depth", "min_samples_split",
                              "min_samples_leaf", "min_weight_fraction_leaf",
                              "max_features", "max_leaf_nodes", "random_state"),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight)

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.scaling = scaling

        # self.estimator_weights = np.random.random((n_estimators,))
        self.estimator_weights = np.ones((n_estimators,))

    def _fit_scale(self, X):
        self.Q5 = []
        self.Q95 = []
        for i in range(X.shape[1]):
            q5, q95 = mquantiles(X[:, i], [0.05, 0.95])
            self.Q5.append(q5)
            self.Q95.append(q95)

    def _scale(self, X):
        X2 = np.copy(X)
        for i in range(X.shape[1]):
            lessidx = np.where(X2[:, i] < self.Q5[i])[0]
            if len(lessidx) > 0:
                lessval = -.01 * np.log(1 + np.log(1 + np.abs(self.Q5[i] - X2[lessidx, i])))
            moreidx = np.where(X2[:, i] > self.Q95[i])[0]
            if len(moreidx) > 0:
                moreval = 1 + .01 * np.log(1 + np.log(1 + np.abs(self.Q5[i] - X2[moreidx, i])))

            denominator = (self.Q95[i] - self.Q5[i]) if self.Q95[i] != self.Q5[i] else 1
            X2[:, i] = (X[:, i] - self.Q5[i]) / denominator

            if len(lessidx) > 0:
                X2[lessidx, i] = lessval
            if len(moreidx) > 0:
                X2[moreidx, i] = moreval

        return X2

    def fit(self, X, y, sample_weight=None):
        if self.scaling:
            self._fit_scale(X)
            super(RRExtraTreesClassifier, self).fit(self._scale(X), y, sample_weight)
        else:
            super(RRExtraTreesClassifier, self).fit(X, y, sample_weight)

    def predict_proba(self, X):
        """Predict class probabilities for X.

        The predicted class probabilities of an input sample is computed as
        the mean predicted class probabilities of the trees in the forest. The
        class probability of a single tree is the fraction of samples of the same
        class in a leaf.

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_samples, n_features]
            The input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csr_matrix``.

        Returns
        -------
        p : array of shape = [n_samples, n_classes], or a list of n_outputs
            such arrays if n_outputs > 1.
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute `classes_`.
        """
        # Check data
        if self.scaling:
            X = self._scale(X)
        X = self._validate_X_predict(X)

        # Assign chunk of trees to jobs
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # Parallel loop
        all_proba = Parallel(n_jobs=n_jobs, verbose=self.verbose,
                             backend="threading")(
            delayed(_parallel_helper)(e, 'predict_proba', X,
                                      check_input=False)
            for e in self.estimators_)

        # Reduce
        proba = all_proba[0]

        if self.n_outputs_ == 1:
            for j in range(1, len(all_proba)):
                proba += self.estimator_weights[j] * all_proba[j]

            # proba /= len(self.estimators_)
            proba /= np.sum(self.estimator_weights[j])

        else:
            for j in range(1, len(all_proba)):
                for k in range(self.n_outputs_):
                    proba[k] += self.estimator_weights[j] * all_proba[j][k]

            for k in range(self.n_outputs_):
                proba[k] /= np.sum(self.estimator_weights[j])

        return proba



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator

from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier, _tree

__author__ = "Romain Ayres, Veltin Dupont, Matthieu Lagacherie, Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class FeatureContributions(BaseEstimator):
    """
    A wrapper for all the action needed to extract the feature importance
    out of the tree interpreter, and plot the results.

        Parameters
        ----------
        model: DecisionTreeRegressor, DecisionTreeClassifier or RandomForestRegressor, RandomForestClassifier
            Scikit-learn model on which the prediction should be decomposed.
            
        estimator_type: {'classification', 'regression' } optional, default 'classification'
            Specify estimator type.
        
        References
        ----------
        [1]  https://github.com/andosa/treeinterpreter
        BSD 3-clause "New" or "Revised" License
    """

    ESTIMATOR_TYPES = {'classification': 0, 'regression': 1}

    def __init__(self, model, estimator_type='classification'):
        self.model = model
        self.Xtest = None
        self.ytest = None
        self.ypred = None
        self.contributions = None

        if estimator_type not in self.ESTIMATOR_TYPES.keys():
            raise ValueError("estimator_type parameter must belong to values:" + self.ESTIMATOR_TYPES.keys())

        self.estimator_type = self.ESTIMATOR_TYPES[estimator_type]

    def predict(self, X):
        """
        Computes the weights given by the tree interpreter.
        
        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Data used by tree interpreter

        Returns
        -------

        """
        # """Computes the weights given by the tree interpreter."""

        predictions, bias, contributions = _TreeInterpreter.predict(self.model, X)
        self._save_predictions(predictions)
        self._format_and_save_contributions(contributions, predictions, X)

        return predictions, bias, contributions

    def _format_and_save_contributions(self, contributions, predictions, dataframe):
        contrib = self._extract_significant_contributions(contributions, predictions)
        formatted_contrib = self._format_contributions_according_to_dataframe_format(contrib, dataframe)
        self._save_contributions(formatted_contrib)

    def _extract_significant_contributions(self, contributions, predictions):
        if self.estimator_type == self.ESTIMATOR_TYPES['classification']:
            contrib = [self._squeeze_class(c, p) for c, p in zip(contributions, predictions)]
        else:
            contrib = contributions
        return contrib

    def _save_predictions(self, predictions):
        self.ypred = predictions

    def _save_contributions(self, contributions):
        self.contributions = contributions

    @staticmethod
    def _format_contributions_according_to_dataframe_format(contributions, dataframe):
        if type(dataframe) == pd.DataFrame:
            new_contributions = pd.DataFrame(contributions, index=dataframe.index, columns=dataframe.columns)
        else:
            new_contributions = contributions
        return new_contributions

    @staticmethod
    def _squeeze_class(contributions, prediction):
        """Reduces the dimension of the tree interpreter output by selecting
        the weight of the predicted category only.
        Ex: if contributions = [[-1, 2, -1],
                                [-2, 5, -3]]
            and prediction = [0.1, 0.9, 0.],
            then it returns [2, 5].

        N.B.: This works well for multi class, but seems dumb for binary classication.

        TODO: implement a fixed column extraction for the binary case.
        """
        pred = np.argmax(prediction)
        return contributions[:, pred]


class _TreeInterpreter:
    @classmethod
    def _get_tree_paths(self, tree, node_id, depth=0):
        """
        Returns all paths through the tree as list of node_ids
        """
        if node_id == _tree.TREE_LEAF:
            raise ValueError("Invalid node_id %s" % _tree.TREE_LEAF)

        left_child = tree.children_left[node_id]
        right_child = tree.children_right[node_id]

        if left_child != _tree.TREE_LEAF:
            left_paths = self._get_tree_paths(tree, left_child, depth=depth + 1)
            right_paths = self._get_tree_paths(tree, right_child, depth=depth + 1)

            for path in left_paths:
                path.append(node_id)
            for path in right_paths:
                path.append(node_id)
            paths = left_paths + right_paths
        else:
            paths = [[node_id]]

        return paths

    @classmethod
    def _predict_tree(self, model, X):
        """
        For a given DecisionTreeRegressor or DecisionTreeClassifier,
        returns a triple of [prediction, bias and feature_contributions], such
        that prediction ≈ bias + feature_contributions.
        """

        leaves = model.apply(X)
        paths = self._get_tree_paths(model.tree_, 0)

        # remove the single-dimensional inner arrays
        values = model.tree_.value.squeeze()

        # check if values have only one axis
        if len(values.shape) == 1:
            values = np.array([values])

        if type(model) == DecisionTreeRegressor:
            line_shape = X.shape[1]
        elif type(model) == DecisionTreeClassifier:
            # scikit stores category counts, we turn them into probabilities
            normalizer = values.sum(axis=1)[:, np.newaxis]
            normalizer[normalizer == 0.0] = 1.0
            values /= normalizer

            line_shape = (X.shape[1], model.n_classes_)

        leaves_set = set(leaves)
        paths_to_compute = [path for path in paths if path[0] in leaves_set]

        contribs = {}
        for _, path in enumerate(paths_to_compute):
            path.reverse()
            leaf = path[-1]
            contribs[leaf] = np.zeros(line_shape)
            for i in range(len(path) - 1):
                contrib = values[path[i + 1]] - \
                          values[path[i]]
                contribs[leaf][model.tree_.feature[path[i]]] += contrib

        contributions = [contribs[leaf] for leaf in leaves]
        direct_prediction = values[leaves]
        biases = [values[path[0]]] * len(X)

        return direct_prediction, biases, contributions

    @classmethod
    def _predict_forest(self, model, X):
        """
        For a given RandomForestRegressor or RandomForestClassifier,
        returns a triple of [prediction, bias and feature_contributions], such
        that prediction ≈ bias + feature_contributions.
        """
        biases = []
        contributions = []
        predictions = []

        for tree in model.estimators_:
            pred, bias, contribution = self._predict_tree(tree, X)
            biases.append(bias)
            contributions.append(contribution)
            predictions.append(pred)
        return (np.mean(predictions, axis=0), np.mean(biases, axis=0),
                np.mean(contributions, axis=0))

    @classmethod
    def predict(self, model, X):
        """ Returns a triple (prediction, bias, feature_contributions), such
        that prediction ≈ bias + feature_contributions.
        
        Parameters
        ----------
        model : DecisionTreeRegressor, DecisionTreeClassifier or
            RandomForestRegressor, RandomForestClassifier
        Scikit-learn model on which the prediction should be decomposed.
        X : array-like, shape = (n_samples, n_features)
        Test samples.
        
        Returns
        -------
        decomposed prediction : triple of
        * prediction, shape = (n_samples) for regression and (n_samples, n_classes)
            for classification
        * bias, shape = (n_samples) for regression and (n_samples, n_classes) for
            classification
        * contributions, shape = (n_samples, n_features) for regression or
            shape = (n_samples, n_features, n_classes) for classification
        """
        # Only single out response variable supported,
        if model.n_outputs_ > 1:
            raise ValueError("Multilabel classification trees not supported")

        if (type(model) == DecisionTreeRegressor or
                    type(model) == DecisionTreeClassifier):
            return self._predict_tree(model, X)
        elif (type(model) == RandomForestRegressor or
                      type(model) == RandomForestClassifier):
            return self._predict_forest(model, X)
        else:
            raise ValueError("Wrong model type. Base learner needs to be \
                DecisionTreeClassifier or DecisionTreeRegressor.")



import os
import unittest
from unittest import skipUnless

import pandas as pd
from pandas.testing import assert_frame_equal
from tests.config import test_user, test_dbname, test_hostname, test_password, test_port

from dsbox.dbconnection.dbconnector import DBconnectorPG


@skipUnless(
    os.getenv('WITH_DB_TESTS'),
    'Please set environment variable WITH_DB_TESTS to enable these tests'
)
class TestDBConnectorPG(unittest.TestCase):

    def test_write_df_to_table_with_to_pg_parameter(self):
        # given
        df = pd.DataFrame({'town': ['Paris', 'New York', 'Roma'],
                           'country': ['France', 'USA', 'Italy'],
                           'trip_cost': [134.6, 234, 85.67]})

        dbconn = DBconnectorPG(username=test_user, password=test_password, hostname=test_hostname, port=test_port,
                               dbname=test_dbname)

        # when
        dbconn.bulk_to_pg(df, 'db.data_town', to_pg_drop=True)

        # then
        self.assertTrue(dbconn.check_table_pg('db', 'data_town'))

    def test_read_table_to_df_from_pg(self):
        # given
        df_true = pd.DataFrame({'town': ['Paris', 'New York', 'Roma'],
                                'country': ['France', 'USA', 'Italy'],
                                'trip_cost': [134.6, 234, 85.67]})
        dbconn = DBconnectorPG(username=test_user, password=test_password, hostname=test_hostname, port=test_port,
                               dbname=test_dbname)

        # when
        df_returned = dbconn.bulk_from_pg('db.data_town')

        # then
        assert_frame_equal(df_true, df_returned)


if __name__ == '__main__':
    unittest.main()



from sqlalchemy import create_engine
import pandas as pd

from dsbox.operators.data_unit import DataOutputUnit, DataInputUnit
from dsbox.dbconnection.dbconnector import DBconnectorPG


class DataInputDBUnit(DataInputUnit):
    """
    Data unit allowing to read a DB table, with SQL engine creation on the fly. It uses Pandas read_sql
    function.

    Parameters
    ----------
    sql_query: str
        SQL query to read data
    db_url: str
        database url
    kwargs: dict
        used by Pandas API to pass others parameters
    """

    def __init__(self, sql_query, db_url, **kwargs):
        self.sql_query = sql_query
        self.db_url = db_url
        self.pandas_kwargs_read = kwargs

    def read_data(self):
        engine = create_engine(self.db_url, echo=False)
        return pd.read_sql(self.sql_query, engine, **self.pandas_kwargs_read)

    def __str__(self):
        return "query: \n{}".format(self.sql_query)


class DataOutputDBUnit(DataOutputUnit):
    """
    Data unit allowing to write a dataframe to a DB table, with SQL engine creation on the fly.
    It uses Pandas to_sql function.

    Parameters
    ----------
    output_name: str
        target SQL table
    db_url: str
        database url
    kwargs: dict
        used by Pandas API to pass others parameters
    """

    def __init__(self, output_name, db_url, **kwargs):
        self.output_name = output_name
        self.db_url = db_url
        self.pandas_kwargs_write = kwargs

    def write_data(self, dataframe):
        engine = create_engine(self.db_url, echo=False)
        dataframe.to_sql(self.output_name, engine, **self.pandas_kwargs_write)

    def __str__(self):
        return self.output_name


class DataPGUnit:
    """
    Global data unit class specilized in PG database operations.

    connection_infos_dict: dict
        dict containing all connection information: username, password, hostname, port, dbname

    """

    def __init__(self, connection_infos_dict):
        self.username = connection_infos_dict['username']
        self.password = connection_infos_dict['password']
        self.hostname = connection_infos_dict['hostname']
        self.port = connection_infos_dict['port']
        self.dbname = connection_infos_dict['dbname']


class DataInputPGUnit(DataPGUnit, DataInputUnit):
    """
    Data unit allowing to read a PG DB table, using efficient bulk operations.

    connection_infos_dict: dict
        dict containing all connection information: username, password, hostname, port, dbname
    """

    def __init__(self, connection_infos_dict, **kwargs):
        super(DataInputPGUnit, self).__init__(connection_infos_dict)
        self.dbconnector_kwargs = kwargs

    def read_data(self):
        dbconnector = DBconnectorPG(self.username, self.password, self.hostname, self.port, self.dbname)
        return dbconnector.bulk_from_pg(**self.dbconnector_kwargs)

    def __str__(self):
        return self.dbconnector_kwargs['table_name']


class DataOutputPGUnit(DataPGUnit, DataOutputUnit):
    """
    Data unit allowing to write a PG DB table, using efficient bulk operations.

    connection_infos_dict: dict
        dict containing all connection information: username, password, hostname, port, dbname
    """

    def __init__(self, connection_infos_dict, **kwargs):
        super(DataOutputPGUnit, self).__init__(connection_infos_dict)
        self.dbconnector_kwargs = kwargs

    def write_data(self, dataframe):
        dbconnector = DBconnectorPG(self.username, self.password, self.hostname, self.port, self.dbname)
        dbconnector.bulk_to_pg(dataframe, **self.dbconnector_kwargs)

    def __str__(self):
        return self.dbconnector_kwargs['table_name']



import gzip
import pickle
import networkx as nx
import matplotlib.pyplot as plt
from smart_open import open

"""
Some util functions used to navigate into Airflow DAGs.
"""


def breadth_first_search_task_list(task_root, task_list=[], mode='upstream'):
    sub_task_list = []
    queue = [task_root]

    while len(queue) > 0:
        task = queue.pop(0)
        sub_task_list.append(task)
        next_tasks = None
        if mode == 'upstream':
            next_tasks = task.upstream_list
        else:
            next_tasks = task.downstream_list
        for next_task in next_tasks:
            if next_task not in queue and next_task not in task_list and next_task not in sub_task_list:
                queue.append(next_task)

    return sub_task_list


def breadth_first_search_shell_list(task_roots):
    shell_task_list = [task_roots]
    done_tasks = set()
    queue = task_roots

    while len(queue) > 0:
        tasks = queue
        next_tasks = []
        for task in tasks:
            for next_task in task.downstream_list:
                if next_task not in done_tasks:
                    next_tasks.append(next_task)
                    done_tasks.add(next_task)

        if len(next_tasks) > 0:
            shell_task_list.append(next_tasks)
        queue = next_tasks

    return shell_task_list


def get_dag_roots(dag):
    roots = []
    for task in dag.tasks:
        if len(task.upstream_list) == 0:
            roots.append(task)

    return roots


def execute_dag(dag, verbose=False, mode='downstream'):
    task_list = []
    roots = dag.roots

    for root in roots:
        sub_task_list = breadth_first_search_task_list(root, task_list, mode=mode)
        task_list = sub_task_list + task_list

    for task in task_list:
        if verbose:
            print(dag.dag_id + '-' + str(task))
        if task.task_type == 'SubDagOperator':
            execute_dag(task.subdag, verbose=verbose)
        else:
            task.execute(dag.get_template_env())

    return task_list


def plot_dag(dag):
    fig, ax = plt.subplots(figsize=(15, 10), dpi=150)

    G = nx.DiGraph()
    color_list = []

    for task in dag.tasks:
        if len(task.downstream_list) > 0:
            for next_task in task.downstream_list:
                G.add_edge(task, next_task)

    for node in G.nodes():
        if len(node.ui_color) == 7:
            color_list.append(node.ui_color)
        else:
            last_code = node.ui_color[-1]
            color_list.append(str(node.ui_color).ljust(7, last_code))

    pos = nx.drawing.nx_agraph.graphviz_layout(G, prog='dot')
    nx.draw_networkx_nodes(G, pos, node_shape='D', node_color=color_list)
    nx.draw_networkx_edges(G, pos, edge_color='gray', alpha=0.8)

    nx.draw_networkx_labels(G, pos, font_size=5)

    ax.set_axis_off()
    plt.title("DAG preview", fontsize=8)
    plt.show()


"""
Some utils function used to persist objects.
"""


def pickle_compress(obj):
    return gzip.zlib.compress(pickle.dumps(obj))


def decompress_unpickle(obj_zp):
    return pickle.loads(gzip.zlib.decompress(obj_zp))


def write_object_file(file_path, obj):
    obj_pz = pickle_compress(obj)
    file_obj = open(file_path, 'wb')
    file_obj.write(obj_pz)
    file_obj.close()


def load_object_file(file_path):
    file_obj = open(file_path, 'rb')
    obj_pz = file_obj.read()
    obj = decompress_unpickle(obj_pz)
    return obj


"""
Some misc utils.
"""


def pandas_downcast_numeric(df_to_downcast, float_type_to_downcast=("float64", "float32"),
                            int_type_to_downcast=("int64", "int32")):
    float_cols = [c for c in df_to_downcast.columns if df_to_downcast[c].dtype == float_type_to_downcast[0]]
    int_cols = [c for c in df_to_downcast.columns if df_to_downcast[c].dtype == int_type_to_downcast[0]]
    df_to_downcast[float_cols] = df_to_downcast[float_cols].apply(lambda x: x.astype(float_type_to_downcast[1]))
    df_to_downcast[int_cols] = df_to_downcast[int_cols].apply(lambda x: x.astype(int_type_to_downcast[1]))


def format_dict_path_items(dictionary, replace_value):
    for k, v in dictionary.items():
        if isinstance(v, dict):
            dictionary[k] = format_dict_path_items(v, replace_value)
        else:
            if isinstance(v, list):
                formatted_list = []
                for list_item in v:
                    if type(list_item) == str:
                        list_item = list_item.format(replace_value)
                    formatted_list.append(list_item)
                dictionary[k] = formatted_list
            else:
                if type(dictionary[k]) == str:
                    dictionary[k] = dictionary[k].format(replace_value)
    return dictionary



from .utils import median_absolute_deviation, mad_outliers, fft_outliers, double_median_absolute_deviation
from .outliers import CovarianceOutliers, GaussianProcessOutliers
from .outliers import KMeansOneClusterOutliers, KMeansIterativeOneClusterOutliers
from .outliers import GMMOutliers, ClusteringOutliers
from .outliers import MADOutliers, FFTOutliers

from . import outliers

__all__ = ["CovarianceOutliers",
           "GaussianProcessOutliers",
           "KMeansOneClusterOutliers",
           "KMeansIterativeOneClusterOutliers",
           "GMMOutliers",
           "ClusteringOutliers",
           "MADOutliers",
           "FFTOutliers"
           ]


import pandas as pd
import numpy as np
import lightgbm as lgb
from dsbox.ml.metrics import root_mean_squared_percentage_error
from dsbox.utils import write_object_file, load_object_file

lgbm_params = {
    'objective':'regression',
    'learning_rate':0.025,
    'n_estimators':1500,
    'max_depth':10
}


def extract_train_dataset(dataframe, min_date=20140101):
    X = dataframe
    X = X[X['Date'] >= min_date]
    X_for_train = X[X['to_predict'] == False]
    X_for_train = X_for_train[X_for_train['MissingValues'] == False]
    X_for_train = X_for_train[X_for_train['Open'] == 1]

    return X_for_train


def train_model(dataframe, features_to_exclude=[], target='Sales', model_path='', min_date=20140101):
    X_for_train = extract_train_dataset(dataframe, min_date=min_date)
    print(X_for_train.shape)

    cols_to_keep = list(set(X_for_train.columns) - set(features_to_exclude))
    print("Amount of features: " + str(len(cols_to_keep)))

    clf_gbm = lgb.LGBMRegressor(**lgbm_params)
    clf_gbm.fit(X_for_train[cols_to_keep], X_for_train[target])

    df_feat_importance = pd.DataFrame({'feature': cols_to_keep, 'importance': clf_gbm.feature_importances_})
    print(df_feat_importance.sort_values('importance', ascending=False))

    print("dumping model")
    write_object_file(model_path + 'lgbm.model', clf_gbm)


def metrics_model(dataframe, features_to_exclude=[], target='Sales', min_date=20140101, date_split=20150615):
    X_for_train = extract_train_dataset(dataframe, min_date=min_date)

    X_train = X_for_train[X_for_train['Date'] < date_split]
    X_test = X_for_train[X_for_train['Date'] >= date_split]

    cols_to_keep = list(set(X_for_train.columns) - set(features_to_exclude))
    print("Amount of features: " + str(len(cols_to_keep)))

    clf_gbm = lgb.LGBMRegressor(**lgbm_params)
    clf_gbm.fit(X_train[cols_to_keep], X_train[target],
                eval_set=[(X_test[cols_to_keep], X_test[target])],
                eval_metric='mse',
                early_stopping_rounds=20,
                verbose=1)

    df_feat_importance = pd.DataFrame({'feature': cols_to_keep, 'importance': clf_gbm.feature_importances_})
    print(df_feat_importance.sort_values('importance', ascending=False))

    y_pred_lgbm = clf_gbm.predict(X_test[cols_to_keep], num_iteration=clf_gbm.best_iteration_)

    score = root_mean_squared_percentage_error(X_test[target], y_pred_lgbm)
    print("RMSPE score: " + str(score))


def predict_model(dataframe, features_to_exclude=[], model_path=''):
    X = dataframe
    X_to_predict = X[X['to_predict'] == True]

    cols_to_keep = list(set(X.columns) - set(features_to_exclude))
    print("Amount of features: " + str(len(cols_to_keep)))

    clf_gbm = load_object_file(model_path + 'lgbm.model')
    y_pred = clf_gbm.predict(X_to_predict[cols_to_keep], num_iteration=clf_gbm.best_iteration_)

    X_to_predict['Sales_pred'] = np.round(y_pred).astype('int')
    X_to_predict['Open'] = X_to_predict['Open'].fillna(0)
    X_to_predict['Sales_pred'] = X_to_predict['Sales_pred'] * X_to_predict['Open']
    X_to_predict['Sales_pred'] = X_to_predict['Sales_pred'].map(lambda x: x if x >= 0 else 0)

    df_to_submit = X_to_predict[['Id', 'Sales_pred']].astype('int')
    df_to_submit['Sales'] = df_to_submit['Sales_pred']
    df_to_submit = df_to_submit.sort_values('Id')
    del df_to_submit['Sales_pred']

    return df_to_submit



import unittest
from datetime import datetime

from airflow import DAG
from airflow.operators.dummy_operator import DummyOperator

from dsbox.utils import execute_dag, format_dict_path_items


class TestUtils(unittest.TestCase):
    def test_execute_dag(self):
        # given
        dag = DAG(dag_id='test_exec_dag', start_date=datetime.now())

        tasks = []
        for i in range(0, 11):
            tasks.append(DummyOperator(task_id='task_' + str(i), dag=dag))

        tasks[0].set_downstream(tasks[1])
        tasks[1].set_downstream(tasks[2])
        tasks[2].set_downstream(tasks[3])
        tasks[2].set_downstream(tasks[4])
        tasks[3].set_downstream(tasks[5])
        tasks[4].set_downstream(tasks[5])
        tasks[4].set_downstream(tasks[8])
        tasks[8].set_downstream(tasks[9])
        tasks[10].set_downstream(tasks[1])

        tasks[6].set_downstream(tasks[7])

        # when
        task_list = execute_dag(dag, mode='downstream')

        # then
        expected_task_lists = [
            [tasks[0], tasks[10], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[8], tasks[9], tasks[6],
             tasks[7]],
            [tasks[10], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[8], tasks[9], tasks[6],
             tasks[7]],

            [tasks[6], tasks[7], tasks[0], tasks[10], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[8],
             tasks[9]],
            [tasks[6], tasks[7], tasks[10], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[8],
             tasks[9]],

            [tasks[0], tasks[10], tasks[1], tasks[2], tasks[3], tasks[4], tasks[8], tasks[9], tasks[5], tasks[6],
             tasks[7]],
            [tasks[10], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[8], tasks[9], tasks[5], tasks[6],
             tasks[7]],

            [tasks[6], tasks[7], tasks[0], tasks[10], tasks[1], tasks[2], tasks[3], tasks[4], tasks[8], tasks[9],
             tasks[5]],
            [tasks[6], tasks[7], tasks[10], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[8], tasks[9],
             tasks[5]],

            [tasks[0], tasks[10], tasks[1], tasks[2], tasks[4], tasks[8], tasks[9], tasks[6], tasks[7], tasks[3],
             tasks[5]],
            [tasks[10], tasks[0], tasks[1], tasks[2], tasks[4], tasks[8], tasks[9], tasks[6], tasks[7], tasks[3],
             tasks[5]],

            [tasks[0], tasks[10], tasks[1], tasks[2], tasks[4], tasks[8], tasks[9], tasks[3], tasks[5], tasks[6],
             tasks[7]],
            [tasks[10], tasks[0], tasks[1], tasks[2], tasks[4], tasks[8], tasks[9], tasks[3], tasks[5], tasks[6],
             tasks[7]],

            [tasks[0], tasks[10], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[6], tasks[7], tasks[8],
             tasks[9]],
            [tasks[10], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[6], tasks[7], tasks[8],
             tasks[9]],

            [tasks[0], tasks[10], tasks[1], tasks[2], tasks[4], tasks[3], tasks[5], tasks[6], tasks[7], tasks[8],
             tasks[9]],
            [tasks[10], tasks[0], tasks[1], tasks[2], tasks[4], tasks[3], tasks[5], tasks[6], tasks[7], tasks[8],
             tasks[9]],

            [tasks[10], tasks[6], tasks[7], tasks[0], tasks[1], tasks[2], tasks[4], tasks[3], tasks[8],  tasks[5],
             tasks[9]],
            [tasks[10], tasks[6], tasks[7], tasks[0], tasks[1], tasks[2], tasks[4], tasks[3], tasks[5], tasks[8],
             tasks[9]],
            [tasks[10], tasks[6], tasks[7], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[8], tasks[5],
             tasks[9]],
            [tasks[10], tasks[6], tasks[7], tasks[0], tasks[1], tasks[2], tasks[3], tasks[4], tasks[5], tasks[8],
             tasks[9]],
        ]

        print(task_list)

        expected_task_list_is_ok = (task_list in expected_task_lists)

        self.assertTrue(expected_task_list_is_ok)

    def test_format_dict_path_items(self):
        # given
        d = {'a': '{}/data',
             'b': {'c': '{}/data1', 'd': '{}/data2'},
             'l': [1, 3, 6],
             'k': ['{}/toto', 8.0],
             'o': False}

        r_value = 'mypath'

        # when
        d_result = format_dict_path_items(d, r_value)

        # then
        d_expected = {'a': 'mypath/data',
                      'b': {'c': 'mypath/data1', 'd': 'mypath/data2'},
                      'l': [1, 3, 6],
                      'k': ['mypath/toto', 8.0],
                      'o': False}

        self.assertDictEqual(d_expected, d_result)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.metrics import accuracy_score

import pyximport
pyximport.install(setup_args={'include_dirs': np.get_include()})

from ._decode import DECODERS
from ._utils import atleast2d_or_csr, safe_sparse_dot, validate_lengths

__credits__ = "Lars Buitinck"


# XXX Should we even derive from ClassifierMixin here?
# We override all the methods.
class BaseSequenceClassifier(BaseEstimator, ClassifierMixin):
    def _get_decoder(self):
        try:
            return DECODERS[self.decode]
        except KeyError:
            raise ValueError("Unknown decoder {0!r}".format(self.decode))

    def predict(self, X, lengths=None):
        """Predict labels/tags for samples X.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Feature matrix.

        lengths : array-like of integer, shape (n_sequences,), optional
            Lengths of sequences in X. If not given, X is assumed to be a
            single sequence of length n_samples.

        Returns
        -------
        y : array, shape (n_samples,)
            Labels per sample in X.
        """
        X = atleast2d_or_csr(X)
        scores = safe_sparse_dot(X, self.coef_.T)
        if hasattr(self, "coef_trans_"):
            n_classes = len(self.classes_)
            coef_t = self.coef_trans_.T.reshape(-1, self.coef_trans_.shape[-1])
            trans_scores = safe_sparse_dot(X, coef_t.T)
            trans_scores = trans_scores.reshape(-1, n_classes, n_classes)
        else:
            trans_scores = None

        decode = self._get_decoder()

        if lengths is None:
            y = decode(scores, trans_scores, self.intercept_trans_,
                       self.intercept_init_, self.intercept_final_)
        else:
            start, end = validate_lengths(X.shape[0], lengths)

            y = [decode(scores[start[i]:end[i]], trans_scores,
                        self.intercept_trans_, self.intercept_init_,
                        self.intercept_final_)
                 for i in range(len(lengths))]
            y = np.hstack(y)

        return self.classes_[y]

    def score(self, X, y, lengths=None):
        """Returns the mean accuracy on the given test data and labels.

        Parameters
        ----------
        X : array-like, shape = (n_samples, n_features)
            Test samples.

        y : array-like, shape = (n_samples,)
            True labels for X.

        lengths : array-like of integer, shape (n_sequences,), optional
            Lengths of sequences in X. If not given, X is assumed to be a
            single sequence of length n_samples.

        Returns
        -------
        score : float
            Mean accuracy of self.predict(X, lengths) wrt. y.
        """
        return accuracy_score(y, self.predict(X, lengths))



import numpy as np
from gensim import corpora
from keras.preprocessing import sequence
from nltk.tokenize import TreebankWordTokenizer
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted, column_or_1d

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class Text2Sequence(BaseEstimator, TransformerMixin):
    """
    Class used to transform text into integers sequence.
    
    Parameters
    ----------
    pad_string : string, optional (default='')
        string value used for padding sequence.
        
    seq_len : int, optional, (default=None)
        fixed sequence length used to build output sequence. If None, this length will be computed during
        transform operation, as the mean of all original sequences + 2 times standard deviation.
        
    tokenizer : object, optional (default=TreebankWordTokenizer())
        object used to tokenize text. The class used must implements a tokenize method.
        
    stemmer : object, optional (default=None)
        object used to stem text. The class used must implements a stem method.
   
   
    Attributes
    ----------
    dictionary_ : gensim.corpora.Dictionary
        maps all tokens to an unique id.
        
    dictionary_size_ : int
        size of the dictionary_ attribute.
            
    Examples
    --------
    >>> import numpy as np
    >>> from nltk.stem.snowball import EnglishStemmer
    >>> from dsbox.ml.neural_networks.processing import Text2Sequence
    >>> X = np.array(['this is really awesome !', \
                      'this is really crap !!'])

    >>> text2seq = Text2Sequence(stemmer=EnglishStemmer())
    >>> text2seq.fit_transform(X)
    array([[0, 5, 3, 4, 2, 1],
           [5, 3, 4, 6, 1, 1]], dtype=int32)
    """

    def __init__(self, pad_string='', tokenizer=TreebankWordTokenizer(), stemmer=None):

        self.pad_string = pad_string
        self.seq_len = None
        self.tokenizer = tokenizer
        self.stemmer = stemmer

        self.attr_to_check = ["dictionary_",
                              "dictionary_size_",
                              ]

    def tokenize(self, text_list):
        """
        Takes a list of texts and return a list of tokens for each text (stemmed or not).
        
        Parameters
        ----------
        text_list : array-like, iterable
            list of strings

        Returns
        -------
            list of token lists

        """
        all_tokens = []
        for text in text_list:
            tokens = self.tokenizer.tokenize(text)
            if self.stemmer is not None:
                tokens = [self.stemmer.stem(word) for word in tokens]
            all_tokens.append(tokens)
        return all_tokens

    def fit(self, X, y=None):
        """
        Fit the transformer by building the word corpus.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str
        y : array-like, shape (n_samples,)
            No used, only here for compatibility reason

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        x = column_or_1d(X)
        all_tokens = self.tokenize(x)
        # adding blank line for considering padding value
        self.dictionary_ = corpora.Dictionary([[self.pad_string]] + all_tokens)

        self.dictionary_size_ = len(self.dictionary_.keys())

        return self

    def transform(self, X):
        """
        Transform a list of texts into sequences of integers (ids), based on a known corpus. If a word is not
        found in the corpus built during fit operation, it will be replaced by the padding string value.
        
        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str
        
        Returns
        -------
        numpy array of ids sequences.
        
        """

        check_is_fitted(self, self.attr_to_check)
        x = column_or_1d(X)

        new_tokens = self.tokenize(x)
        pad_value = self.dictionary_.token2id[self.pad_string]

        # transform all tokens into unique ids
        word_ids, word_ids_len = [], []
        for doc in new_tokens:
            word_id = []
            for word in doc:
                if word in self.dictionary_.token2id:
                    word_id.append(self.dictionary_.token2id[word])
                else:
                    word_id.append(pad_value)

            word_ids.append(word_id)
            word_ids_len.append(len(word_id))

        # compute the length sequence
        if self.seq_len is None:
            self.seq_len = np.round((np.mean(word_ids_len) + 2 * np.std(word_ids_len))).astype(int)

        return sequence.pad_sequences(np.array(word_ids), maxlen=self.seq_len, value=pad_value)



__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


def shinglelize(text, shingle_size=3):
    """
    Transform an input string to a list of shingles (or n-grams) with a given size.

    Parameters
    ----------
    text : string
        Text to shinglelize

    shingle_size: int, optional, default 3
        Size of a shingle

    Returns
    -------
    shingles : list
        List of shingles
    """

    shingles = []
    i = 0
    while (i + (shingle_size - 1)) < len(text):
        shingles.append(text[i:(i + shingle_size)])
        i = i + 1

    return shingles



import logging
import unittest

import numpy as np

from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.python.keras.utils.np_utils import to_categorical

logging.getLogger("tensorflow").setLevel(logging.WARNING)

from dsbox.ml.neural_networks.keras_factory.text_models import LSTMFactory

np.random.seed(42)


def dummy_hash_function(x):
    code = 0
    for c in x:
        code += ord(c)
    return code


class TestKerasTextModels(unittest.TestCase):
    def test_LSTM_compilation_and_fit_predict_without_execution_error(self):
        # given
        x_train = np.array(['this is really awesome !',
                            'this is really crap !!']
                           )
        y_train = np.array([1, 0])

        ids_x_train = np.empty([2, 5])
        for i in range(0, len(x_train)):
            ids = [dummy_hash_function(token) for token in x_train[i].split(' ')]
            ids_x_train[i, :] = ids
        num_labels = 2
        y_enc = to_categorical(y_train, num_labels)
        dictionary_size = np.int(np.max(ids_x_train) + 1)

        # when
        lstm_factory = LSTMFactory()
        clf_keras = KerasClassifier(build_fn=lstm_factory.create_model,
                                    dictionary_size=dictionary_size,
                                    num_labels=num_labels)
        clf_keras.fit(ids_x_train, y_enc, epochs=1, verbose=False)

        x_test = np.array(['it is really awesome !'])
        ids_x_test = np.empty([1, 5])
        ids_x_test[0, :] = [dummy_hash_function(token) for token in x_test[0].split(' ')]

        y_pred = clf_keras.predict(ids_x_test)

        # then
        self.assertIsNotNone(y_pred)


if __name__ == '__main__':
    unittest.main()



class DataExecutor:
    """
    Same behavior as the Data Operator, without Airflow dependency. Based on Data Units, it can be used to execute
    operation directly by calling the execute() method.

    Parameters
    ----------
    operation_function: function
        function taking the output of the input_unit and returning its result to the input of the output_unit
    input_unit: DataInputUnit, default=None
        define input data reading
    output_unit: DataOutputUnit, default=None
        define output data writing
    """

    def __init__(self, operation_function, input_unit=None, output_unit=None, **kwargs):
        self.operation_function = operation_function
        self.input_unit = input_unit
        self.output_unit = output_unit
        self.kwargs = kwargs

    def execute(self):
        data = None
        if self.input_unit is not None:
            data = self.input_unit.read_data()
            data = self.operation_function(data, **self.kwargs)
        else:
            data = self.operation_function(**self.kwargs)
        if self.output_unit is not None:
            self.output_unit.write_data(data)



import importlib

from dsbox.ml.neural_networks import KerasFactory
from dsbox.ml.neural_networks.keras_factory.image_models import KerasApplicationFactory
from dsbox.ml.neural_networks.processing import Text2Sequence
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils.validation import check_is_fitted
import tensorflow as tf
from tensorflow import keras
from tensorflow.python.keras.utils.np_utils import to_categorical
import numpy as np

__author__ = "Vincent Levorato"
__credits__ = "https://github.com/octo-technology/bdacore"
__license__ = "Apache 2.0"


class TextNeuralNetPipeline(BaseEstimator, ClassifierMixin):
    """
    Pipeline used to fit/predict raw text in a classification problem context. It transforms the text data
    via Text2Sequence transformer and passes the result to a keras model.
    
    Parameters
    ----------
    text2seq : object (default=Text2Sequence())
        transformer used to transform text into integer sequences.
        
    factory_class : KerasFactory (default=KerasFactory)
        set the factory used to build the keras model. By default, it takes an abstract class, which has to 
        be replaced by a concrete one (LSTMFactory or CNN_LSTMFactory for instance).
        
    num_labels : int, optional (default=2)
        amount of class labels
        
    Attributes
    ----------
    model_ : KerasClassifier
        keras scikit wrapper containing original keras architecture model.
    
            
    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(42) # for making doctests stable
    >>> from dsbox.ml.neural_networks.keras_factory.text_models import LSTMFactory
    >>> from dsbox.ml.neural_networks.processing.workflow import TextNeuralNetPipeline
    
    >>> x_train = np.array([u'this is really really awesome !', \
                            u'it is so awesome !', \
                            u'that sucks'] \
                           )
    >>> y_train = np.array([1, 1, 0])

    >>> model = TextNeuralNetPipeline(factory_class=LSTMFactory, num_labels=2, random_state=42)
    >>> _ = model.fit(x_train, y_train, verbose=0) # to avoid doctests error

    >>> x_test = np.array([u'it is really awesome !'])
    >>> model.predict(x_test)
    array([1])
    
    """

    def __init__(self, text2seq=Text2Sequence(), factory_class=KerasFactory, num_labels=2, random_state=42):
        self.text2seq = text2seq
        self.factory = factory_class()
        self.num_labels = num_labels
        self.random_state = random_state

        self.attr_to_check = ["model_"]

    def fit(self, X, y, **kwargs):
        """
        Fit the workflow by building the word corpus, and fitting the keras model.
    
        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str
        y : array-like, shape (n_samples,)
            Class targets.
        **kwargs : 
            parameters passed to inner keras model
    
        Returns
        -------
        self : object
            Returns an instance of self.
        """

        x = self.text2seq.fit_transform(X)
        y_enc = to_categorical(y, self.num_labels)

        self.model_ = KerasClassifier(build_fn=self.factory.create_model,
                                      dictionary_size=self.text2seq.dictionary_size_,
                                      num_labels=self.num_labels)

        self.model_.fit(x, y_enc, **kwargs)

        return self

    def predict(self, X):
        """
        Predict a list of texts to belong to a known class.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        numpy array of class ids.

        """
        check_is_fitted(self, self.attr_to_check)

        x = self.text2seq.transform(X)
        return self.model_.predict(x)

    def predict_proba(self, X):
        """
        Predict a list of texts to belong to a known class.

        Parameters
        ----------
        X : array-like, iterable
            Collection of str or an iterable which yields str

        Returns
        -------
        probability of belonging to each class id.
        """

        check_is_fitted(self, self.attr_to_check)

        x = self.text2seq.transform(X)
        return self.model_.predict_proba(x)


class ImageNeuralNetPipeline(BaseEstimator, ClassifierMixin):
    """
    Pipeline used to predict an image class. It uses pre-trained models proposed by Keras applications package.

    Parameters
        ----------
        module_name: str, optional (default='tensorflow.keras.applications.xception')
            specify keras application module name to import model
        model_name: str, optional (default='Xception')
            specify keras application model name to use
        top_pred: int, optional (default=1)
            parameter used to select amount of image categories returned by the predict method
        **kwargs :
            parameters passed to inner keras model

    .. Keras Grad-CAM example: https://keras.io/examples/vision/grad_cam/
    """
    def __init__(self,
                 module_name='tensorflow.keras.applications.xception', model_name='Xception',
                 top_pred=1,
                 **kwargs
                 ):
        self.factory = KerasApplicationFactory()
        self.module_name = module_name
        self.model_name = model_name
        self.top_pred = top_pred

        self.preprocess_input = getattr(importlib.import_module(module_name), 'preprocess_input')
        self.decode_predictions = getattr(importlib.import_module(module_name), 'decode_predictions')

        self.last_conv_layer_name_ = None
        self.classifier_layer_names_ = None
        self.img_size_ = None

        self.attr_to_check = ["model_",
                              "last_conv_layer_name_",
                              "classifier_layer_names_"]
        self.kwargs = kwargs

    def _get_img_array(self, img_path, size):
        img = keras.preprocessing.image.load_img(img_path, target_size=size)
        array = keras.preprocessing.image.img_to_array(img)
        array = np.expand_dims(array, axis=0)
        return array

    def _make_gradcam_heatmap(self, img_array, model, last_conv_layer_name, classifier_layer_names):
        # First, we create a model that maps the input image to the activations
        # of the last conv layer
        last_conv_layer = model.get_layer(last_conv_layer_name)
        last_conv_layer_model = keras.Model(model.inputs, last_conv_layer.output)

        # Second, we create a model that maps the activations of the last conv
        # layer to the final class predictions
        classifier_input = keras.Input(shape=last_conv_layer.output.shape[1:])
        x = classifier_input
        for layer_name in classifier_layer_names:
            x = model.get_layer(layer_name)(x)
        classifier_model = keras.Model(classifier_input, x)

        # Then, we compute the gradient of the top predicted class for our input image
        # with respect to the activations of the last conv layer
        with tf.GradientTape() as tape:
            # Compute activations of the last conv layer and make the tape watch it
            last_conv_layer_output = last_conv_layer_model(img_array)
            tape.watch(last_conv_layer_output)
            # Compute class predictions
            preds = classifier_model(last_conv_layer_output)
            top_pred_index = tf.argmax(preds[0])
            top_class_channel = preds[:, top_pred_index]

        # This is the gradient of the top predicted class with regard to
        # the output feature map of the last conv layer
        grads = tape.gradient(top_class_channel, last_conv_layer_output)

        # This is a vector where each entry is the mean intensity of the gradient
        # over a specific feature map channel
        pooled_grads = tf.reduce_mean(grads, axis=(0, 1, 2))

        # We multiply each channel in the feature map array
        # by "how important this channel is" with regard to the top predicted class
        last_conv_layer_output = last_conv_layer_output.numpy()[0]
        pooled_grads = pooled_grads.numpy()
        for i in range(pooled_grads.shape[-1]):
            last_conv_layer_output[:, :, i] *= pooled_grads[i]

        # The channel-wise mean of the resulting feature map
        # is our heatmap of class activation
        heatmap = np.mean(last_conv_layer_output, axis=-1)

        # For visualization purpose, we will also normalize the heatmap between 0 & 1
        heatmap = np.maximum(heatmap, 0) / np.max(heatmap)
        return heatmap

    def get_heatmap(self, x):
        """
        Return a matrix representing the heatmap of the image to understand model prediction assumptions.
        Parameters
        ----------
        x: str
            path to image file

        Returns
        -------
        NumPy heatmap matrix

        """
        check_is_fitted(self, self.attr_to_check)

        img_array = self.preprocess_input(self._get_img_array(x, size=self.img_size_))
        return self._make_gradcam_heatmap(img_array, self.model_, self.last_conv_layer_name_,
                                                   self.classifier_layer_names_)

    def fit(self, X=None, y=None, **kwargs):
        """
        Build pre-trained model by extracting layers information.

        Parameters
        ----------
        X : not used, present for API consistence purpose
        y : not used, present for API consistence purpose
        **kwargs :
            parameters passed to inner keras model

        Returns
        -------
        self : object
            Returns an instance of self.
        """

        self.model_ = self.factory.create_model(module_name=self.module_name,
                                                model_name=self.model_name,
                                                **self.kwargs)

        input_size = self.model_.get_config()['layers'][0]['config']['batch_input_shape']
        self.img_size_ = (input_size[1], input_size[2])

        layers = self.model_.get_config()['layers']
        last_conv_layer_name = None
        last_conv_layer_index = -1
        i = len(layers) - 1
        while last_conv_layer_name is None and i >= 0:
            if 'conv' in layers[i]['name']:
                last_conv_layer_name = layers[i]['name']
                last_conv_layer_index = i
            i -= 1

        self.last_conv_layer_name_ = last_conv_layer_name

        classifier_layer_names = []
        for i in range(last_conv_layer_index + 1, len(layers)):
            classifier_layer_names.append(layers[i]['name'])

        self.classifier_layer_names_ = classifier_layer_names

        return self

    def predict(self, X):
        """
        Predict image tags.

        Parameters
        ----------
        X : array-like, iterable
            Collection of image paths or an iterable which yields image path

        Returns
        -------
        numpy array of predicted image label with probability

        """
        check_is_fitted(self, self.attr_to_check)
        predictions = []
        for x in X:
            img_array = self.preprocess_input(self._get_img_array(x, size=self.img_size_))
            preds = self.model_.predict(img_array)
            predictions.append(self.decode_predictions(preds, top=self.top_pred)[0])

        return predictions



import unittest
import numpy as np

from dsbox.ml.markov.hmmlearn import hmm
from tests.ml.markov.hmmlearn import log_likelihood_increasing, make_covar_matrix, normalized


class GaussianHMMTestMixin(unittest.TestCase):
    covariance_type = 'diag'  # set by subclasses

    def setUp(self):
        self.prng = prng = np.random.RandomState(10)
        self.n_components = n_components = 3
        self.n_features = n_features = 3
        self.startprob = prng.rand(n_components)
        self.startprob = self.startprob / self.startprob.sum()
        self.transmat = prng.rand(n_components, n_components)
        self.transmat /= np.tile(self.transmat.sum(axis=1)[:, np.newaxis],
                                 (1, n_components))
        self.means = prng.randint(-20, 20, (n_components, n_features))
        self.covars = make_covar_matrix(
            self.covariance_type, n_components, n_features, random_state=prng
        )

    def test_bad_covariance_type(self):
        self.setUp()
        with self.assertRaises(ValueError):
            h = hmm.GaussianHMM(20, covariance_type='badcovariance_type')
            h.means_ = self.means
            h.covars_ = []
            h.startprob_ = self.startprob
            h.transmat_ = self.transmat
            h._check()

    def test_score_samples_and_decode(self):
        self.setUp()
        h = hmm.GaussianHMM(self.n_components, self.covariance_type,
                            init_params="st")
        h.means_ = self.means
        h.covars_ = self.covars

        # Make sure the means are far apart so posteriors.argmax()
        # picks the actual component used to generate the observations.
        h.means_ = 20 * h.means_

        gaussidx = np.repeat(np.arange(self.n_components), 5)
        n_samples = len(gaussidx)
        X = self.prng.randn(n_samples, self.n_features) + h.means_[gaussidx]
        h._init(X)
        ll, posteriors = h.score_samples(X)

        self.assertEqual(posteriors.shape, (n_samples, self.n_components))
        assert np.allclose(posteriors.sum(axis=1), np.ones(n_samples))

        viterbi_ll, stateseq = h.decode(X)
        assert np.allclose(stateseq, gaussidx)

    def test_sample(self, n=1000):
        self.setUp()
        h = hmm.GaussianHMM(self.n_components, self.covariance_type)
        h.startprob_ = self.startprob
        h.transmat_ = self.transmat
        # Make sure the means are far apart so posteriors.argmax()
        # picks the actual component used to generate the observations.
        h.means_ = 20 * self.means
        h.covars_ = np.maximum(self.covars, 0.1)

        X, state_sequence = h.sample(n, random_state=self.prng)
        self.assertEqual(X.shape, (n, self.n_features))
        self.assertEqual(len(state_sequence), n)

    def test_fit(self, params='stmc', n_iter=5, **kwargs):
        self.setUp()
        h = hmm.GaussianHMM(self.n_components, self.covariance_type)
        h.startprob_ = self.startprob
        h.transmat_ = normalized(
            self.transmat + np.diag(self.prng.rand(self.n_components)), 1)
        h.means_ = 20 * self.means
        h.covars_ = self.covars

        lengths = [10] * 10
        X, _state_sequence = h.sample(sum(lengths), random_state=self.prng)

        # Mess up the parameters and see if we can re-learn them.
        # TODO: change the params and uncomment the check
        self.assertIsNotNone(h.fit(X, lengths=lengths))
        # assert log_likelihood_increasing(h, X, lengths, n_iter)

    def test_fit_sequences_of_different_length(self):
        self.setUp()
        lengths = [3, 4, 5]
        X = self.prng.rand(sum(lengths), self.n_features)

        h = hmm.GaussianHMM(self.n_components, self.covariance_type)
        # This shouldn't raise
        # ValueError: setting an array element with a sequence.
        self.assertIsNotNone(h.fit(X, lengths=lengths))

    def test_fit_with_length_one_signal(self):
        self.setUp()
        lengths = [10, 8, 1]
        X = self.prng.rand(sum(lengths), self.n_features)

        h = hmm.GaussianHMM(self.n_components, self.covariance_type)
        # This shouldn't raise
        # ValueError: zero-size array to reduction operation maximum which
        #             has no identity
        self.assertIsNotNone(h.fit(X, lengths=lengths))

    def test_fit_zero_variance(self):
        self.setUp()
        # Example from issue #2 on GitHub.
        X = np.asarray([
            [7.15000000e+02, 5.85000000e+02, 0.00000000e+00, 0.00000000e+00],
            [7.15000000e+02, 5.20000000e+02, 1.04705811e+00, -6.03696289e+01],
            [7.15000000e+02, 4.55000000e+02, 7.20886230e-01, -5.27055664e+01],
            [7.15000000e+02, 3.90000000e+02, -4.57946777e-01, -7.80605469e+01],
            [7.15000000e+02, 3.25000000e+02, -6.43127441e+00, -5.59954834e+01],
            [7.15000000e+02, 2.60000000e+02, -2.90063477e+00, -7.80220947e+01],
            [7.15000000e+02, 1.95000000e+02, 8.45532227e+00, -7.03294373e+01],
            [7.15000000e+02, 1.30000000e+02, 4.09387207e+00, -5.83621216e+01],
            [7.15000000e+02, 6.50000000e+01, -1.21667480e+00, -4.48131409e+01]
        ])

        h = hmm.GaussianHMM(3, self.covariance_type)
        self.assertIsNotNone(h.fit(X))

    def test_fit_with_priors(self, params='stmc', n_iter=5):
        self.setUp()
        startprob_prior = 10 * self.startprob + 2.0
        transmat_prior = 10 * self.transmat + 2.0
        means_prior = self.means
        means_weight = 2.0
        covars_weight = 2.0
        if self.covariance_type in ('full', 'tied'):
            covars_weight += self.n_features
        covars_prior = self.covars

        h = hmm.GaussianHMM(self.n_components, self.covariance_type)
        h.startprob_ = self.startprob
        h.startprob_prior = startprob_prior
        h.transmat_ = normalized(
            self.transmat + np.diag(self.prng.rand(self.n_components)), 1)
        h.transmat_prior = transmat_prior
        h.means_ = 20 * self.means
        h.means_prior = means_prior
        h.means_weight = means_weight
        h.covars_ = self.covars
        h.covars_prior = covars_prior
        h.covars_weight = covars_weight

        lengths = [200] * 10
        X, _state_sequence = h.sample(sum(lengths), random_state=self.prng)

        # Re-initialize the parameters and check that we can converge to the
        # original parameter values.
        h_learn = hmm.GaussianHMM(self.n_components, self.covariance_type,
                                  params=params)
        h_learn.n_iter = 0
        h_learn.fit(X, lengths=lengths)

        self.assertTrue(log_likelihood_increasing(h_learn, X, lengths, n_iter))

        # Make sure we've converged to the right parameters.
        # a) means
        self.assertTrue(np.allclose(sorted(h.means_.tolist()),
                                    sorted(h_learn.means_.tolist()),
                                    0.01))
        # b) covars are hard to estimate precisely from a relatively small
        #    sample, thus the large threshold
        self.assertTrue(np.allclose(sorted(h._covars_.tolist()),
                                    sorted(h_learn._covars_.tolist()),
                                    10))


class TestGaussianHMMWithSphericalCovars(GaussianHMMTestMixin, unittest.TestCase):
    covariance_type = 'spherical'

    def test_fit_startprob_and_transmat(self):
        self.test_fit('st')


class TestGaussianHMMWithDiagonalCovars(GaussianHMMTestMixin, unittest.TestCase):
    covariance_type = 'diag'

    def test_covar_is_writeable(self):
        h = hmm.GaussianHMM(n_components=1, covariance_type="diag",
                            init_params="c")
        X = np.random.normal(size=(1000, 5))
        h._init(X)

        # np.diag returns a read-only view of the array in NumPy 1.9.X.
        # Make sure this doesn't prevent us from fitting an HMM with
        # diagonal covariance matrix. See PR#44 on GitHub for details
        # and discussion.
        self.assertTrue(h._covars_.flags["WRITEABLE"])

    def test_fit_left_right(self):
        transmat = np.zeros((self.n_components, self.n_components))

        # Left-to-right: each state is connected to itself and its
        # direct successor.
        for i in range(self.n_components):
            if i == self.n_components - 1:
                transmat[i, i] = 1.0
            else:
                transmat[i, i] = transmat[i, i + 1] = 0.5

        # Always start in first state
        startprob = np.zeros(self.n_components)
        startprob[0] = 1.0

        lengths = [10, 8, 1]
        X = self.prng.rand(sum(lengths), self.n_features)

        h = hmm.GaussianHMM(self.n_components, covariance_type="diag",
                            params="mct", init_params="cm")
        h.startprob_ = startprob.copy()
        h.transmat_ = transmat.copy()
        h.fit(X)

        self.assertTrue((h.startprob_[startprob == 0.0] == 0.0).all())
        self.assertTrue((h.transmat_[transmat == 0.0] == 0.0).all())

        posteriors = h.predict_proba(X)
        self.assertFalse(np.isnan(posteriors).any())
        self.assertTrue(np.allclose(posteriors.sum(axis=1), 1.))

        score, state_sequence = h.decode(X, algorithm="viterbi")
        self.assertTrue(np.isfinite(score))


class TestGaussianHMMWithTiedCovars(GaussianHMMTestMixin, unittest.TestCase):
    covariance_type = 'tied'


class TestGaussianHMMWithFullCovars(GaussianHMMTestMixin, unittest.TestCase):
    covariance_type = 'full'



import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer

from pyproj import Proj, transform

from dsbox.ml.feature_engineering import TagEncoder
from dsbox.utils import write_object_file, load_object_file


def lambert_coord_to_long_lat(x, y, inProj, outProj):
    return transform(inProj, outProj, x, y)


def transform_df_coordinates(dataframe, x_column, y_column):
    inProj = Proj(init="epsg:3945")
    outProj = Proj(proj='lonlat', ellps='WGS84')

    dataframe[['long', 'lat']] = dataframe.apply(lambda row: lambert_coord_to_long_lat(row[x_column], row[y_column],
                                                                                       inProj, outProj), axis=1) \
        .apply(pd.Series)

    return dataframe


def join_dataframes(dataframe_list, **kwargs):
    X = dataframe_list[0]
    for dataframe in dataframe_list[1:len(dataframe_list)]:
        X = X.join(dataframe, **kwargs)

    return X


def dict_missing_feature(dataframe, missing_values_feature, groupby_feature, dummy_feature):
    df = dataframe.groupby([groupby_feature, missing_values_feature])[[dummy_feature]].count().reset_index(drop=False)
    df.columns = [groupby_feature, missing_values_feature, 'count']
    df_count = df.groupby(groupby_feature)[['count']].max().reset_index(drop=False)
    df = df.merge(df_count, on=[groupby_feature, 'count'])
    mapping_dict = df[[groupby_feature, missing_values_feature]].set_index(groupby_feature).to_dict()[
        missing_values_feature]

    return mapping_dict


def notediag_to_num(text):
    if text == "Arbre d'avenir normal":
        return 0
    if text == "Arbre d'avenir incertain":
        return 1
    if "dans les 10 ans" in text:
        return 2
    if "dans les 5 ans" in text:
        return 3
    if "diatement" in text:
        return 4

    return -1


def prio_renouv_to_num(text):
    if text == "plus de 20 ans":
        return 20
    if "de 11" in text:
        return 11
    if "de 6" in text:
        return 6
    if "de 1 " in text:
        return 1

    return -1


def remarques_to_num(text):
    if pd.isnull(text) or text == '0':
        return 0

    if "mort" in text:
        return 2

    return 1


def fillna_columns(dataframe, simple_features=[], mode='train', model_path=None):
    X = dataframe

    if mode == 'train':
        imputer_ANNEEREALISATIONDIAGNOSTIC = SimpleImputer(missing_values=np.nan, strategy='mean')
        imputer_ANNEEREALISATIONDIAGNOSTIC.fit(X[['ANNEEREALISATIONDIAGNOSTIC']])
        write_object_file(model_path + 'imputer_ANNEEREALISATIONDIAGNOSTIC.feat', imputer_ANNEEREALISATIONDIAGNOSTIC)
    else:
        imputer_ANNEEREALISATIONDIAGNOSTIC = load_object_file(model_path + 'imputer_ANNEEREALISATIONDIAGNOSTIC.feat')

    X['ANNEEREALISATIONDIAGNOSTIC'] = imputer_ANNEEREALISATIONDIAGNOSTIC.transform(X[['ANNEEREALISATIONDIAGNOSTIC']])
    X['ANNEEREALISATIONDIAGNOSTIC'] = np.round(X['ANNEEREALISATIONDIAGNOSTIC']).astype('int')

    X['ANNEETRAVAUXPRECONISESDIAG'] = X['ANNEETRAVAUXPRECONISESDIAG'].fillna(-1)

    miss_group_features = [('ESPECE', 'GENRE_BOTA'),
                           ('DIAMETREARBREAUNMETRE', 'ESPECE'),
                           ('VARIETE', 'ESPECE')
                           ]
    for tuple in miss_group_features:

        missing_feature = tuple[0]
        grouping_feature = tuple[1]

        if mode == 'train':
            mapping_dict = dict_missing_feature(X, missing_feature, grouping_feature, 'CODE')
            write_object_file(model_path + 'mapping_dict_' + missing_feature + '.feat', mapping_dict)
        else:
            mapping_dict = load_object_file(model_path + 'mapping_dict_' + missing_feature + '.feat')

        grouping_feature_values = X[grouping_feature].unique()
        for grouping_feature_value in grouping_feature_values:
            if grouping_feature_value not in mapping_dict:
                mapping_dict[grouping_feature_value] = 'inconnu'

        X[missing_feature] = X.apply(
            lambda row: mapping_dict[row[grouping_feature]] if pd.isnull(row[missing_feature]) else row[
                missing_feature], axis=1)

    for feature in simple_features:
        X[feature] = X[feature].fillna('inconnu')

    return X


def category_to_numerical_features(dataframe, features, mode='train', model_path=None):
    X = dataframe

    for feature in features:
        if mode == 'train':
            tagencoder = TagEncoder()
            X[feature] = tagencoder.fit_transform(X[feature])
            write_object_file(model_path + 'tagencoder_' + feature + '.feat', tagencoder)
        else:
            tagencoder = load_object_file(model_path + 'tagencoder_' + feature + '.feat')
            X[feature] = tagencoder.transform(X[feature])

    X['DIAMETREARBREAUNMETRE'] = X['DIAMETREARBREAUNMETRE'].map(lambda x: '-1' if x == 'inconnu' else x)
    X['DIAMETREARBREAUNMETRE'] = X['DIAMETREARBREAUNMETRE'].map(lambda x: x.split(' ')[0]).astype('int')

    X['NOTEDIAGNOSTIC'] = X['NOTEDIAGNOSTIC'].apply(notediag_to_num)
    X['PRIORITEDERENOUVELLEMENT'] = X['PRIORITEDERENOUVELLEMENT'].apply(prio_renouv_to_num)
    X['REMARQUES'] = X['REMARQUES'].apply(remarques_to_num)

    return X



import random


class FilenameGenerator:
    """
    Class used to auto-generate temporary filenames used by intermediate operations. It adds an hexadecimal tag
    to a chosen prefix, like for instance 'X_1acf'. It keeps in memory all the generated hextags.

    Parameters
    ----------
    path: str
        temporary files path
    filename: str, default='X'
        prefix to temporary filenames
    hextag_length: int, default=4
        number of hexa digits used to generate file names
    sepchar: str, default='_'
        char separator between prefix filenames and hexa digits
    seed: int, default=123
        seed used by pseudo-random number generator

    Attributes
    ----------
    hextags: list
        contains all generated hexa tags

    """
    def __init__(self, path='', filename='X', hextag_length=4, sepchar='_', seed=123):
        self.hextags = []
        self.path = path
        self.filename = filename
        self.hextag_length = hextag_length
        self.sepchar = sepchar
        self.seed = seed
        random.seed(seed)

        self.max_val = int('F' * hextag_length, base=16)

    def generate_hextag(self):
        return hex(random.randint(0, self.max_val))[2:].rjust(self.hextag_length, '0')

    def generate_filename(self, pos=None):
        """
        Generates a random filename and keeps track of it by adding its hextag to hextags list

        Parameters
        ----------
        pos: int, default=None
            if None, the function generates a hextag, adds it to hextags list and returns generated filename.
            if not None, the function returns the previous generated filename with the pos index from the hextags list,
            but if pos gives an out of bounds index, it has the same behavior as when set to None.

        Returns
        -------
        :str
            generated filename path
        """


        if pos is None:
            hextag = self.generate_hextag()
            self.hextags.append(hextag)
        else:
            if pos >= len(self.hextags):
                for i in range(len(self.hextags), pos + 1):
                    self.hextags.append(self.generate_hextag())
            hextag = self.hextags[pos]

        return self.path + self.filename + self.sepchar + hextag



import numpy as np
import unittest

from scipy.special import logsumexp
from dsbox.ml.markov.hmmlearn.base import ConvergenceMonitor, _BaseHMM


class TestMonitor(unittest.TestCase):
    def test_converged_by_iterations(self):
        m = ConvergenceMonitor(tol=1e-3, n_iter=2, verbose=False)
        self.assertFalse(m.converged)
        m.report(-0.01)
        self.assertFalse(m.converged)
        m.report(-0.1)
        self.assertTrue(m.converged)

    def test_converged_by_logprob(self):
        m = ConvergenceMonitor(tol=1e-3, n_iter=10, verbose=False)
        for logprob in [-0.03, -0.02, -0.01]:
            m.report(logprob)
            self.assertFalse(m.converged)

        m.report(-0.0101)
        self.assertTrue(m.converged)

    def test_reset(self):
        m = ConvergenceMonitor(tol=1e-3, n_iter=10, verbose=False)
        m.iter = 1
        m.history.append(-0.01)
        m._reset()
        self.assertEqual(m.iter, 0)
        self.assertFalse(m.history)


class StubHMM(_BaseHMM):
    """An HMM with hardcoded observation probabilities."""

    def _compute_log_likelihood(self, X):
        return self.framelogprob


class TestBaseAgainstWikipedia(unittest.TestCase):
    def setup_method(self):
        # Example from http://en.wikipedia.org/wiki/Forward-backward_algorithm
        self.framelogprob = np.log([[0.9, 0.2],
                                    [0.9, 0.2],
                                    [0.1, 0.8],
                                    [0.9, 0.2],
                                    [0.9, 0.2]])

        h = StubHMM(2)
        h.transmat_ = [[0.7, 0.3], [0.3, 0.7]]
        h.startprob_ = [0.5, 0.5]
        h.framelogprob = self.framelogprob
        self.hmm = h

    def test_do_forward_pass(self):
        self.setup_method()
        logprob, fwdlattice = self.hmm._do_forward_pass(self.framelogprob)

        reflogprob = -3.3725
        self.assertEqual(np.round(logprob, 4), reflogprob)
        reffwdlattice = np.array([[0.4500, 0.1000],
                                  [0.3105, 0.0410],
                                  [0.0230, 0.0975],
                                  [0.0408, 0.0150],
                                  [0.0298, 0.0046]])
        self.assertTrue(np.allclose(np.exp(fwdlattice), reffwdlattice, 4))

    def test_do_backward_pass(self):
        self.setup_method()
        bwdlattice = self.hmm._do_backward_pass(self.framelogprob)

        refbwdlattice = np.array([[0.0661, 0.0455],
                                  [0.0906, 0.1503],
                                  [0.4593, 0.2437],
                                  [0.6900, 0.4100],
                                  [1.0000, 1.0000]])
        self.assertTrue(np.allclose(np.exp(bwdlattice), refbwdlattice, 4))

    def test_do_viterbi_pass(self):
        self.setup_method()
        logprob, state_sequence = self.hmm._do_viterbi_pass(self.framelogprob)

        refstate_sequence = [0, 0, 1, 0, 0]
        self.assertTrue(np.allclose(state_sequence, refstate_sequence))

        reflogprob = -4.4590
        self.assertEqual(np.round(logprob, 4), reflogprob)

    def test_score_samples(self):
        self.setup_method()
        # ``StubHMM` ignores the values in ```X``, so we just pass in an
        # array of the appropriate shape.
        logprob, posteriors = self.hmm.score_samples(self.framelogprob)
        self.assertTrue(np.allclose(posteriors.sum(axis=1), np.ones(len(posteriors))))

        reflogprob = -3.3725
        self.assertEqual(np.round(logprob, 4), reflogprob)

        refposteriors = np.array([[0.8673, 0.1327],
                                  [0.8204, 0.1796],
                                  [0.3075, 0.6925],
                                  [0.8204, 0.1796],
                                  [0.8673, 0.1327]])
        self.assertTrue(np.allclose(posteriors, refposteriors, atol=1e-4))


class TestBaseConsistentWithGMM(unittest.TestCase):
    def setup_method(self):
        self.n_components = 8
        n_samples = 10

        self.framelogprob = np.log(np.random.random((n_samples, self.n_components)))

        h = StubHMM(self.n_components)
        h.framelogprob = self.framelogprob

        # If startprob and transmat are uniform across all states (the
        # default), the transitions are uninformative - the model
        # reduces to a GMM with uniform mixing weights (in terms of
        # posteriors, not likelihoods).
        h.startprob_ = np.ones(self.n_components) / self.n_components
        h.transmat_ = np.ones((self.n_components, self.n_components)) / self.n_components

        self.hmm = h

    def test_score_samples(self):
        self.setup_method()
        logprob, hmmposteriors = self.hmm.score_samples(self.framelogprob)

        n_samples, n_components = self.framelogprob.shape
        self.assertTrue(np.allclose(hmmposteriors.sum(axis=1), np.ones(n_samples)))

        norm = logsumexp(self.framelogprob, axis=1)[:, np.newaxis]
        gmmposteriors = np.exp(self.framelogprob - np.tile(norm, (1, self.n_components)))
        self.assertTrue(np.allclose(hmmposteriors, gmmposteriors))

    def test_decode(self):
        self.setup_method()
        _logprob, state_sequence = self.hmm.decode(self.framelogprob)

        n_samples, n_components = self.framelogprob.shape
        norm = logsumexp(self.framelogprob, axis=1)[:, np.newaxis]
        gmmposteriors = np.exp(self.framelogprob -
                               np.tile(norm, (1, n_components)))
        gmmstate_sequence = gmmposteriors.argmax(axis=1)
        self.assertTrue(np.allclose(state_sequence, gmmstate_sequence))


class TestHMMAttributes(unittest.TestCase):
    def test_base_hmm_attributes(self):
        n_components = 20
        startprob = np.random.random(n_components)
        startprob /= startprob.sum()
        transmat = np.random.random((n_components, n_components))
        transmat /= np.tile(transmat.sum(axis=1)[:, np.newaxis], (1, n_components))

        h = StubHMM(n_components)

        self.assertEqual(h.n_components, n_components)

        h.startprob_ = startprob
        self.assertTrue(np.allclose(h.startprob_, startprob))

        with self.assertRaises(ValueError):
            h.startprob_ = 2 * startprob
            h._check()
        with self.assertRaises(ValueError):
            h.startprob_ = []
            h._check()
        with self.assertRaises(ValueError):
            h.startprob_ = np.zeros((n_components - 2, 2))
            h._check()

        h.startprob_ = startprob
        h.transmat_ = transmat
        self.assertTrue(np.allclose(h.transmat_, transmat))
        with self.assertRaises(ValueError):
            h.transmat_ = 2 * transmat
            h._check()
        with self.assertRaises(ValueError):
            h.transmat_ = []
            h._check()
        with self.assertRaises(ValueError):
            h.transmat_ = np.zeros((n_components - 2, n_components))
            h._check()


class TestStationaryDistribution(unittest.TestCase):
    def test_stationary_distribution(self):
        n_components = 10
        h = StubHMM(n_components)
        transmat = np.random.random((n_components, n_components))
        transmat /= np.tile(transmat.sum(axis=1)[:, np.newaxis], (1, n_components))
        h.transmat_ = transmat
        stationary = h.get_stationary_distribution()
        self.assertEqual(stationary.dtype, float)
        self.assertListEqual(list(np.round(np.dot(h.get_stationary_distribution().T, h.transmat_), 5)),
                             list(np.round(stationary, 5)))



import logging
import unittest

import numpy as np
from tensorflow.keras.datasets import mnist

from dsbox.ml.neural_networks.keras_factory.autoencoders import DeepAutoEncoderFactory, AutoEncoderClassifier

logging.getLogger("tensorflow").setLevel(logging.WARNING)

np.random.seed(42)


class TestKerasTAutoEncoderModels(unittest.TestCase):
    def test_DeepAE_compilation_and_fit_predict_without_execution_error(self):
        # given
        (x_train, _), (x_test, _) = mnist.load_data()
        x_train = x_train[:100].astype('float32') / 255.
        x_test = x_test[:20].astype('float32') / 255.
        x_train = x_train.reshape((len(x_train), np.prod(x_train.shape[1:])))
        x_test = x_test.reshape((len(x_test), np.prod(x_test.shape[1:])))

        # when
        ae_factory = DeepAutoEncoderFactory()
        autoencoder, encoder = ae_factory.create_model(start_layer_size=256, sub_layers_level=4)
        clf_keras = AutoEncoderClassifier(autoencoder, encoder)

        clf_keras.fit(x_train, x_train,
                      epochs=2,
                      batch_size=256,
                      shuffle=True,
                      validation_data=(x_test, x_test),
                      verbose=False)

        y_pred = clf_keras.predict(x_test)

        # then
        self.assertIsNotNone(y_pred)


if __name__ == '__main__':
    unittest.main()
